import os

import unittest
from pyoma.inference.tabledef import BestMatchesFmt, GenomeFmt
import pyoma.inference.db
import tables
import numpy as np
import tempfile


def get_test_data():
    return (
        np.array(
            [
                (1, 1, 251.32, 22.4, 9.11, 422, True, False, True),
                (1, 5, 2251.32, 2.4, 3.11, 852, True, True, True),
                (1, 6, 151.32, 122.4, 59.11, 223, False, False, False),
                (2, 4, 211.55, 33.1, 10.1, 599, True, True, True),
                (2, 8, 1111.52, 13.1, 10.1, 799, False, False, True),
                (5, 6, 222, 31.7332, 222.1, 554, True, True, True),
            ],
            dtype=tables.dtype_from_descr(BestMatchesFmt),
        ),
        np.array([0, 3, 3, 5, 5, 5, 5, 5, 5, 6]).reshape(5, 2),
    )


class RelationsOfEntryTest(unittest.TestCase):
    def setUp(self):
        data, idx = get_test_data()
        self.indata = data[np.where(data["EntryNr1"] == 1)]

    def test_contains(self):
        all_pairs = pyoma.inference.db.RelationsOfEntry(self.indata)
        self.assertFalse(4 in all_pairs)
        for en2 in (1, 5, 6):
            self.assertTrue(en2 in all_pairs, str(en2) + " is not in all_pairs")

        spairs = pyoma.inference.db.StablePairsOfEntry(self.indata)
        self.assertFalse(6 in spairs)
        for en2 in (1, 5):
            self.assertTrue(en2 in spairs)

        vpairs = pyoma.inference.db.VPairsOfEntry(self.indata)
        for en2 in (2, 1, 6):
            self.assertFalse(en2 in vpairs)
        self.assertTrue(5 in vpairs)

    def test_iterator(self):
        spairs = pyoma.inference.db.StablePairsOfEntry(self.indata)
        self.assertEqual([x["EntryNr2"] for x in spairs], [1, 5], "Unexpected members in SPairs")

    def test_matches(self):
        vpairs = pyoma.inference.db.VPairsOfEntry(self.indata)
        self.assertEqual(1, len(vpairs.relations()), "Unexpected number of VPs returned")


class GenomePairTest(unittest.TestCase):
    def setUp(self):
        matches, itab = get_test_data()
        itab = np.array([0, 3, 3, 5, 5, 5, 5, 5, 5, 6]).reshape(5, 2)
        self.genome_pair = pyoma.inference.db.GenomePair(matches, itab)

    def test_get_entry(self):
        prot1_matches = [(m["EntryNr1"], m["EntryNr2"]) for m in self.genome_pair[1]["ALL"]]
        self.assertEqual(
            [(1, 1), (1, 5), (1, 6)],
            prot1_matches,
            "unexpected matches with ALL selector",
        )
        prot2_matches = [(m["EntryNr1"], m["EntryNr2"]) for m in self.genome_pair[2]["SP"]]
        self.assertEqual([(2, 4)], prot2_matches, "unexpected matches with SP selector")
        self.assertEqual(
            0,
            len(self.genome_pair[3]["ALL"].relations()),
            "Unexpected matches, should be empty",
        )

    def test_keep_different_reltypes_in_sync(self):
        prot1_before = [z["EntryNr2"] for z in self.genome_pair[1]["BM"]]
        self.assertEqual([1, 5], prot1_before, "test-setup failed")
        self.genome_pair[1]["VP"] = [1, 6]
        prot2_after = [z["EntryNr2"] for z in self.genome_pair[1]["BM"]]
        self.assertEqual([1, 5, 6], prot2_after, "change of VP should set BM flag")

    def test_fail_if_not_member(self):
        self.assertRaises(ValueError, self.genome_pair[2]["BM"].set_relations, [4, 6, 8])

    def test_iterators_in_sync(self):
        # create iterator before any changes on data
        bm_it = iter(self.genome_pair[1]["BM"])
        # set new bm pairs
        new_matches = [5, 6]
        self.genome_pair[1]["BM"] = new_matches
        matches_from_iterator = [z["EntryNr2"] for z in bm_it]
        self.assertEqual(matches_from_iterator, new_matches)


class OmaDBTest(unittest.TestCase):
    def setUp(self):
        matches, itab = get_test_data()
        genome_summary = np.array(
            [
                (1234, "SE001", 5, 21122, 0, "Synthetic Genome 1", "Test"),
                (1235, "SE002", 8, 15222, 5, "Synthetic Genome 2", "Test"),
            ],
            tables.dtype_from_descr(GenomeFmt),
        )
        self.db_filename = tempfile.mktemp(".h5")
        fd = tables.open_file(self.db_filename, "w")
        se1vs2 = fd.create_group("/Matches/SE001", "SE002", createparents=True)
        fd.create_table(se1vs2, "Relations", BestMatchesFmt, obj=matches)
        fd.create_carray(se1vs2, "ProteinIndex", obj=itab)
        fd.create_table("/", "GenomeSummary", GenomeFmt, obj=genome_summary)
        fd.close()

    def tearDown(self):
        try:
            os.remove(self.db_filename)
        except IOError:
            pass

    def test_without_write_sps(self):
        db = pyoma.inference.db.OmaDB(self.db_filename, mode="a")
        pair = db.matches("SE001", "SE002")
        sp = [z["EntryNr2"] for z in pair[1]["SP"]]
        sp.append(6)
        pair[1]["SP"] = sp
        sp_after = [z["EntryNr2"] for z in pair[1]["SP"]]
        self.assertEqual(sp, sp_after)
        db.close()

    def test_write_vps(self):
        db = pyoma.inference.db.OmaDB(self.db_filename, mode="a")
        vp = [z["EntryNr2"] for z in db.matches("SE001", "SE002")[1]["VP"]]
        vp.append(6)
        db.matches("SE001", "SE002")[1]["VP"] = vp
        db.matches("SE001", "SE002").flush()
        db.close()

        db = pyoma.inference.db.OmaDB(self.db_filename, mode="r")
        vp = [z["EntryNr2"] for z in db.matches("SE001", "SE002")[1]["VP"]]
        self.assertIn(6, vp)
        db.close()



import ete3
import numpy as np
import pandas as pd


def generate_treeweights(tree: ete3.PhyloNode, taxa_index):
    # weighing function for tax level, masking levels etc. sets all weights to 1
    """
    Generate the weights of each taxonomic level to be applied during the
    constructin of weighted minhashes
    :param tree: trimmed species tree
    :param taxa_index: dict mapping taxa to columns
    :return: weights: a vector of weights for each tax level
    """

    weights = np.zeros((3 * len(taxa_index), 1))
    for i in range(3):
        for n in tree.traverse():
            weights[len(taxa_index) * i + taxa_index[n.name]] = 1
    return weights


def hash_tree_profile(tp: ete3.PhyloNode, taxa_index, species_index, treeweights, wmg):
    """
    Generate a weighted minhash and binary matrix row for a tree profile

    :param tp: a pyham tree profile
    :param taxa_index: dict mapping taxa to columns
    :param treeweights: a vector of weights for each tax levels
    :param wmg: Datasketch weighted minhash generator
    :return hog_matrix: a vector of weights for each tax level
    :return weighted_hash: a weighted minhash of a HOG

    """

    losses = [taxa_index[n.name] for n in tp.traverse() if n.lost and n.name in taxa_index]
    dupl = [taxa_index[n.name] for n in tp.traverse() if n.dupl and n.name in taxa_index]
    presence = [taxa_index[n.name] for n in tp.traverse() if n.nbr_genes > 0 and n.name in taxa_index]
    covered_species = [species_index[n.name] for n in tp.iter_leaves() if n.nbr_genes > 0 and n.name in species_index]
    indices = dict(zip(["presence", "loss", "dup"], [presence, losses, dupl]))
    hog_matrix_weighted = np.zeros((1, 3 * len(taxa_index)))
    hog_matrix_binary = np.zeros((1, 3 * len(taxa_index)))
    species_matrix = np.zeros(len(species_index))
    species_matrix[covered_species] = 1
    for i, event in enumerate(indices):
        if len(indices[event]) > 0:
            hogindex = np.asarray(indices[event]) + i * len(taxa_index)
            hog_matrix_weighted[:, hogindex] = treeweights[hogindex, :].ravel()
            hog_matrix_binary[:, hogindex] = 1
    weighted_hash = wmg.minhash(list(hog_matrix_weighted.flatten()))
    return hog_matrix_binary, weighted_hash, species_matrix


def row2hash(row, taxa_index, species_index, treeweights, wmg):
    """
    turn a dataframe row with an orthoxml file to hash and matrix row
    :param row: lsh builder dataframe row (with family, treeprofile)
    :param taxa_index: dict mapping taxa to columnsfam
    :param treeweights: a vector of weights for each tax levels
    :param wmg: Datasketch weighted minhash generator
    :return: hog_matrix: a vector of weights for each tax level
    :return: weighted_hash: a weighted minhash of a HOG
    """
    # convert a dataframe row to a weighted minhash
    fam, treemap = row.tolist()
    hog_matrix, weighted_hash, species_matrix = hash_tree_profile(treemap, taxa_index, species_index, treeweights, wmg)
    return pd.Series([weighted_hash, hog_matrix, species_matrix], index=["hash", "rows", "species"])



import collections
import logging
import os.path
import re
import numpy
from typing import Union

logger = logging.getLogger(__name__)


class HogLevelFilter(object):
    """Filter sub-hogs at given level (or below)

    This class allows to identify  at a given level or - if a hog
    does not go deep enough to cover the desired level, the whole hog will be returned.
    """

    def __init__(self, db, level):
        self.db = db
        self.level = level
        tax_below_level = self.db.tax.get_subtaxonomy_rooted_at(level)
        self.sub_level_distances = self._build_distance_to_query_level_lookup(tax_below_level)
        self.parent_levels = set(
            db.tax.get_parent_taxa(db.tax.get_taxnode_from_name_or_taxid(level)[0]["NCBITaxonId"])["Name"]
        )
        # cache all the families nrs that are defined in the clade of interest
        self.fams_in_clade = set(
            int(roothog["Fam"])
            for roothog in db.get_hdf5_handle().root.HogLevel.where('~contains(ID, b".") & (IsRoot == True)')
            if roothog["Level"] in self.sub_level_distances or roothog["Level"] in self.parent_levels
        )

    def _build_distance_to_query_level_lookup(self, tax):
        nodes = {}

        def traverse(node, dist=0):
            nodes[node["name"].encode("utf-8")] = dist
            try:
                for c in node["children"]:
                    traverse(c, dist + 1)
            except KeyError:
                pass

        traverse(tax.as_dict())
        return nodes

    def find_closest_children_level(self, levs):
        logger.debug(levs)
        min_dist, closest = 1000, None

        for lev in levs:
            try:
                if self.sub_level_distances[lev] < min_dist:
                    closest = lev.item().decode()
                    min_dist = self.sub_level_distances[lev]
            except KeyError:
                pass
        logger.debug("closest level %s at %s", closest, min_dist)
        return closest

    def analyse_families(self, it):
        for fam in it:
            if fam not in self.fams_in_clade:
                continue
            subhog_ids = self.db.get_subhogids_at_level(fam_nr=fam, level=self.level)
            if len(subhog_ids) == 0:
                # let's check if we have any level in the family that is a sub clade
                levs = self.db.hog_levels_of_fam(fam_nr=fam)
                closest_level = self.find_closest_children_level(levs)
                if closest_level is not None:
                    yield (self.db.format_hogid(fam), closest_level)
            else:
                for subhog_id in subhog_ids:
                    yield (subhog_id, self.level)


def build_hog_to_og_map(db):
    hog_to_og_cnts = collections.defaultdict(collections.Counter)
    entries = db.get_hdf5_handle().get_node("/Protein/Entries")
    for protein in entries.where("(OmaHOG != b'') & (OmaGroup > 0)"):
        hog_id = protein["OmaHOG"]
        og = int(protein["OmaGroup"])
        for p in re.finditer(rb"\.", hog_id):
            cur_id = hog_id[: p.start()]
            hog_to_og_cnts[cur_id].update((og,))
        hog_to_og_cnts[hog_id].update((og,))
    for hog_id, cnts in hog_to_og_cnts.items():
        yield hog_id, cnts.most_common(1)


def compare_levels(
    parent_level_hogs: numpy.array,
    children_level_hogs: numpy.array,
    return_duplication_events=False,
):
    """compares hogs at two levels and returns duplicated/lost/gained/identical states

    The function requires that the hogs are sorted numpy arrays according to their HOGid"""
    annotated = []
    dups = 0
    i = j = 0
    while i < len(parent_level_hogs) and j < len(children_level_hogs):
        if parent_level_hogs[i]["Fam"] < children_level_hogs[j]["Fam"]:
            annotated.append(tuple(parent_level_hogs[i]) + (b"lost",))
            i += 1
        elif parent_level_hogs[i]["Fam"] > children_level_hogs[j]["Fam"]:
            annotated.append(tuple(children_level_hogs[j]) + (b"gained",))
            j += 1
        else:
            if parent_level_hogs[i]["ID"] == children_level_hogs[j]["ID"]:
                annotated.append(tuple(children_level_hogs[j]) + (b"retained",))
                i += 1
                j += 1
            else:
                dupl_fnd = 0
                while j < len(children_level_hogs) and children_level_hogs[j]["ID"].startswith(
                    parent_level_hogs[i]["ID"]
                ):
                    annotated.append(tuple(children_level_hogs[j]) + (b"duplicated",))
                    j += 1
                    dupl_fnd += 1
                if dupl_fnd > 0:
                    # parent duplicated into at least one subhog
                    dups += 1
                else:
                    # parent subhog does not exist, it is lost
                    annotated.append(tuple(parent_level_hogs[i]) + (b"lost",))
                i += 1

    while i < len(parent_level_hogs):
        annotated.append(tuple(parent_level_hogs[i]) + (b"lost",))
        i += 1
    while j < len(children_level_hogs):
        annotated.append(tuple(children_level_hogs[j]) + (b"gained",))
        j += 1
    dt = children_level_hogs.dtype.descr + [("Event", "S10")]
    diff = numpy.array(annotated, dtype=dt)
    if return_duplication_events:
        return diff, dups
    return diff


hog_re = re.compile(r"(?P<id>HOG:(?P<rel>[A-Z]+)?(?P<fam>\d+)(?P<sub>[a-z0-9.]*))(?:_(?P<taxid>\d+))?")
sub_re = re.compile(r"^(?P<nr>\d+)(?P<char>[a-z]+)$")


def are_orthologous(
    hogid1: Union[bytes, str], hogid2: Union[bytes, str], return_group: bool = False
) -> Union[bool, str]:
    h1 = hogid1 if isinstance(hogid1, str) else hogid1.decode()
    h2 = hogid2 if isinstance(hogid2, str) else hogid2.decode()

    m1 = hog_re.match(h1)
    m2 = hog_re.match(h2)
    if m1.group("fam") != m2.group("fam") or m1.group("rel") != m2.group("rel"):
        return False
    common_group = ["HOG:{}{}".format(m1.group("rel"), m1.group("fam"))]
    for sub_i, sub_j in zip(m1.group("sub").split("."), m2.group("sub").split(".")):
        if sub_i == sub_j:
            if len(sub_i) > 0:
                common_group.append(sub_i)
            continue
        m_sub_i = sub_re.match(sub_i)
        m_sub_j = sub_re.match(sub_j)
        if m_sub_i.group("nr") != m_sub_j.group("nr"):
            if return_group:
                return ".".join(common_group)
            else:
                return True
        elif m_sub_i.group("char") != m_sub_j.group("char"):
            return False
    if return_group:
        return ".".join(common_group)
    else:
        return True



import numpy
import collections
from lark import Lark, Transformer, ParseError
from tables import dtype_from_descr
from .tablefmt import LocusTable
import logging


logger = logging.getLogger(__name__)

"""This package is intended to parse the darwin locus structure
and create a numpy recarray out of it. """


Exon = collections.namedtuple("Exon", ["start", "end", "strand"])

grammar = """?locus : join | complement | complement_join | location
             join  : "join" "(" (complement | location ) ("," (complement | location ))+ ")"
             complement : "complement(" location ")"
             complement_join : "complement" "(" "join" "(" location ("," location)+ ")" ")"
             location : pos [ _RNG pos ] | "FromElsewhere" "('" _SEQID "'," pos [ _RNG pos ] ")"
             ?pos : num | "Before" "(" num ")" | "After" "(" num ")"
             ?num : /[0-9]+/             -> number
             _SEQID: /[A-Za-z0-9._-]+/
             _RNG: ".."


             %import common.NUMBER
             %import common.WS
             %ignore WS"""


class LocusTransformer(Transformer):
    def number(self, vals):
        return int(vals[0])

    def location(self, value):
        return Exon(value[0], value[1] if len(value) > 1 else value[0], 1)

    def complement(self, value):
        rev = [e._replace(strand=-1 * e.strand) for e in value]
        if len(rev) == 1:
            return rev[0]
        else:
            return rev

    def complement_join(self, value):
        return self.complement(value)

    def join(self, values):
        return values


class LocusParser(object):
    def __init__(self):
        self.parser = Lark(grammar, start="locus")
        self.locus_transformer = LocusTransformer()
        self.dtype = dtype_from_descr(LocusTable)

    def parse(self, locus_string, entry_nr=0):
        try:
            tree = self.parser.parse(locus_string)
        except ParseError as e:
            raise ValueError("cannot parse '{}' locus string".format(locus_string))
        data = self.locus_transformer.transform(tree)
        nr_exons = 1 if isinstance(data, Exon) else len(data)
        locus_data = numpy.empty(nr_exons, dtype=self.dtype)
        locus_data[["Start", "End", "Strand"]] = data
        locus_data["EntryNr"] = entry_nr
        return locus_data



import os
import tempfile
import unittest
import tempfile
import gzip
import bz2
from pyoma.common import auto_open


class AutoOpenBaseTest(unittest.TestCase):
    filesuffix = None

    def setUp(self) -> None:
        self.expected_text = """This is a test text. Let's see if we can properly load it."""
        with tempfile.NamedTemporaryFile(suffix=self.filesuffix, delete=False) as fh:
            self.testfilename = fh.name

    def tearDown(self) -> None:
        os.remove(self.testfilename)


class AutoOpenRegularReadTest(AutoOpenBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.store_text(self.testfilename)

    def store_text(self, fn):
        with open(fn, mode="wt", encoding="utf-8") as fh:
            fh.write(self.expected_text)

    def test_read_as_text(self):
        with auto_open(self.testfilename, "rt") as fh:
            res = fh.read()
        self.assertEqual(res, self.expected_text)

    def test_read_as_bytes(self):
        with auto_open(self.testfilename, "rb") as fh:
            res = fh.read()
        self.assertEqual(res, self.expected_text.encode("utf-8"))


class AutoOpenGzipTest(AutoOpenRegularReadTest):
    def store_text(self, fn):
        with gzip.open(fn, mode="wt", encoding="utf-8") as fh:
            fh.write(self.expected_text)


class AutoOpenBz2Test(AutoOpenRegularReadTest):
    def store_text(self, fn):
        with bz2.open(fn, mode="wt", encoding="utf-8") as fh:
            fh.write(self.expected_text)


class AutoOpenRegularWriteTest(AutoOpenBaseTest):
    def expected_text_start(self):
        return self.expected_text.encode("utf-8")

    def test_write(self):
        with auto_open(self.testfilename, "wt") as fh:
            fh.write(self.expected_text)
        with open(self.testfilename, "rb") as fh:
            res = fh.read()
            self.assertTrue(res.startswith(self.expected_text_start()))


class AutoOpenGzipWriteTest(AutoOpenRegularWriteTest):
    filesuffix = ".gz"

    def expected_text_start(self):
        return b"\x1f\x8b\x08"


class AutoOpenBz2WriteTest(AutoOpenRegularWriteTest):
    filesuffix = ".bz2"

    def expected_text_start(self):
        return b"\x42\x5a\x68"


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, print_function, division

import csv
import gzip
import io
import json
import os
import shutil
import tempfile
import unittest
import logging
from hashlib import md5

import numpy
import tables
import pyoma.browser.tablefmt as tablefmt
from pyoma.browser.convert import (
    callDarwinExport,
    DarwinExporter,
    compute_ortholog_types,
    load_tsv_to_numpy,
    HogConverter,
)
from .test_db import find_path_to_test_db


def store_in_json(data, fn):
    os.mkdir(os.path.dirname(fn))
    with open(fn, "w") as fd:
        json.dump(data, fd)


class ImportDummyBase(unittest.TestCase):
    """This base class makes sure that the environment vars
    `DARWIN_BROWSERDATA_PATH` and `DARWIN_NETWORK_SCRATCH_PATH`
    are set to a temporary path at the beginning of the test cases
    and get restored to the initial value after a test cases
    finished.

    At the beginning of each testcase, a new DarwinExporter is
    created at the location of $DARWIN_BROWSERDATA_PATH and removed
    after the testcase.

    Subclasses can therefor overwrite the $DARWIN_BROWSERDATA_PATH
    in the setUpClass method if needed.
    """

    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.old_env = {(z, os.getenv(z, None)) for z in ("DARWIN_NETWORK_SCRATCH_PATH", "DARWIN_BROWSERDATA_PATH")}
        os.environ["DARWIN_NETWORK_SCRATCH_PATH"] = cls.tmpdir
        os.environ["DARWIN_BROWSERDATA_PATH"] = cls.tmpdir

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)
        for var, val in cls.old_env:
            if val is None:
                os.unsetenv(var)
            else:
                os.environ[var] = val

    def setUp(self):
        fname = tempfile.mktemp(".h5", "OmaServer-", self.tmpdir)
        self.darwin_exporter = DarwinExporter(fname)

    def tearDown(self):
        fn = self.darwin_exporter.h5.filename
        self.darwin_exporter.close()
        os.remove(fn)


class ImportIntegrationBase(ImportDummyBase):
    """This base class checks if the Test-case data
    is available on any of the groups shared directories
    and uses this directory for the DARWIN_BROWSERDATA_PATH.
    If none of the shares is available, the tests are skipped.
    """

    @classmethod
    def setUpClass(cls):
        super(ImportIntegrationBase, cls).setUpClass()
        test_data_available = False
        for folder in (
            "/work/FAC/FBM/DBC/cdessim2/oma/oma-browser",
            "/cs/research/biosciences/oma/oma-server",
            "/mnt/oma-browser",
            "/Volumes/oma-browser",
        ):
            if os.path.isdir(folder):
                test_data_available = True
                break
        if not test_data_available:
            raise unittest.SkipTest("data not available")
        os.environ["DARWIN_BROWSERDATA_PATH"] = os.path.join(folder, "Test.Jul2014", "data")


class GenomeDirectImportTest(ImportIntegrationBase):
    def compare_genomes_tab(self, data):
        self.darwin_exporter.add_species_data()
        gstab = self.darwin_exporter.h5.get_node("/Genome")
        self.assertEqual(len(gstab), len(data["GS"]), "unexpected number of genomes")
        for genome in data["GS"]:
            gs = gstab.read_where("UniProtSpeciesCode == code", genome[1].encode("utf-8"))
            for key in (
                (2, "TotEntries"),
                (3, "TotAA"),
                (0, "NCBITaxonId"),
                (5, "SciName"),
            ):
                expected = genome[key[0]]
                if isinstance(expected, str):
                    expected = expected.encode("utf-8")
                self.assertEqual(
                    gs[key[1]],
                    expected,
                    "data doesn't match for {}: {} vs {}".format(key[0], gs[key[1]], expected),
                )
        taxtab = self.darwin_exporter.h5.get_node("/Taxonomy")
        all_taxlevels = taxtab[:]
        self.assertFalse(
            numpy.where(all_taxlevels["NCBITaxonId"] == all_taxlevels["ParentTaxonId"])[0].any(),
            "must not have taxlevel pointing to itself",
        )

    def test_load_species_from_json(self):
        data = self.darwin_exporter.call_darwin_export("GetGenomeData();")
        json_fname = os.path.join(self.tmpdir, "pyoma", "gs.json")
        store_in_json(data, json_fname)
        self.compare_genomes_tab(data)
        os.remove(json_fname)

    def test_load_species_from_darwin(self):
        data = self.darwin_exporter.call_darwin_export("GetGenomeData();")
        self.compare_genomes_tab(data)


class ProteinImportViaJson(ImportIntegrationBase):
    @classmethod
    def setUpClass(cls):
        super(ProteinImportViaJson, cls).setUpClass()
        data = callDarwinExport("GetGenomeData();")
        json_fname = os.path.join(cls.tmpdir, "pyoma", "gs.json")
        store_in_json(data, json_fname)

    def test_add_proteins(self):
        self.darwin_exporter.add_species_data()
        self.darwin_exporter.add_proteins()

        entry_tab = self.darwin_exporter.h5.get_node("/Protein/Entries")
        sequence_tab = self.darwin_exporter.h5.get_node("/Protein/SequenceBuffer")
        for i, e in enumerate(entry_tab):
            self.assertEqual(
                i + 1,
                e["EntryNr"],
                "entries are not ordered: {} - {}".format(i, e["EntryNr"]),
            )
            seq = sequence_tab[e["SeqBufferOffset"] : e["SeqBufferOffset"] + e["SeqBufferLength"] - 1].tostring()
            self.assertEqual(
                md5(seq).hexdigest(),
                e["MD5ProteinHash"].decode(),
                "sequence hashes disagree for {}".format(e["EntryNr"]),
            )

    def test_get_version(self):
        version = self.darwin_exporter.get_version()
        self.assertIn("Test", version)
        self.darwin_exporter.add_version()
        self.assertEqual(version, self.darwin_exporter.h5.get_node_attr("/", "oma_version"))

    def test_add_orthologs_from_darwin(self):
        pass


class OrthologyTypeTester(unittest.TestCase):
    def setUp(self):
        self._tmpdir = tempfile.mkdtemp()
        self.exp = DarwinExporter(os.path.join(self._tmpdir, "test.h5"))
        self.pw = self.exp.h5.create_table("/", "VPairs", tablefmt.PairwiseRelationTable)

    def test_convert_to_numpytable(self):
        rels = [[1, 5], (2, 6)]
        tab = self.exp._convert_to_numpyarray(rels, self.pw)
        self.assertEqual(len(tab), 2)
        # test that default value are used
        self.assertTrue(numpy.all(tab["Distance"] == -1))

    def test_orthology_type_inference(self):
        genome_offs = numpy.array([0, 4, 9, 19])
        rels = [
            [1, 5],
            [1, 6],
            [1, 10],
            [1, 20],
            [2, 5],
            [2, 6],
            [2, 10],
            [2, 11],
            [3, 12],
            [3, 21],
            [3, 22],
        ]
        expected_rel_type = numpy.array([3, 3, 2, 0, 3, 3, 3, 1, 0, 1, 1])

        reltab = self.exp._convert_to_numpyarray(rels, self.pw)
        compute_ortholog_types(reltab, genome_offs)
        self.assertTrue(numpy.array_equal(reltab["RelType"], expected_rel_type))

    def tearDown(self):
        self.exp.close()
        shutil.rmtree(self._tmpdir)


class TSVOrthologyFileTester(unittest.TestCase):
    def setUp(self):
        with tempfile.NamedTemporaryFile(suffix=".gz", delete=False) as fh:
            self.datafile = fh.name

    def tearDown(self):
        try:
            os.remove(self.datafile)
        except OSError:
            pass

    def put_data_in_file(self, data):
        with gzip.open(self.datafile, mode="wt", compresslevel=6) as cmp:
            writer = csv.writer(cmp, delimiter=str("\t"))
            for row in data:
                writer.writerow(row)

    def test_regular_example(self):
        data = [
            (1, 1, 12421, "1:1", 0.99, 1.523),
            (2, 5, 21214, "n:1", 0.94, 12.14),
            (3, 5, 23552, "n:1", 0.91, 21.2),
        ]
        self.put_data_in_file(data)
        res = load_tsv_to_numpy((self.datafile, 0, 100, False))
        self.assertEqual(3, res.size)
        self.assertTrue((numpy.arange(1, 4) == res["EntryNr1"]).all())
        self.assertTrue((numpy.array((101, 105, 105)) == res["EntryNr2"]).all())
        self.assertTrue((numpy.array((0, 2, 2)) == res["RelType"]).all())
        self.assertTrue(((res["Distance"] > 0) & (res["Distance"] < 22)).all())
        self.assertTrue(((res["AlignmentOverlap"] > 0.9) & (res["AlignmentOverlap"] <= 1)).all())

    def test_empty_file(self):
        self.put_data_in_file([])
        res = load_tsv_to_numpy((self.datafile, 0, 0, False))
        self.assertEqual(0, res.size)


class H5HelpersTests(ImportDummyBase):
    @staticmethod
    def get_table_data():
        return numpy.array([(1, 9), (2, 11), (5, 21)], dtype=[("X", "i4"), ("Y", "u2")])

    def setUp(self):
        super(H5HelpersTests, self).setUp()
        data = self.get_table_data()
        self.darwin_exporter.h5.create_table("/", "Example", obj=data)

    def test_create_table_if_needed_without_existing_table(self):
        self.darwin_exporter.create_table_if_needed("/", "XRef2", description=self.get_table_data().dtype)
        self.assertEqual(0, len(self.darwin_exporter.h5.get_node("/XRef2")))

    def test_create_table_if_needed_not_needed(self):
        self.darwin_exporter.create_table_if_needed("/", "Example")
        self.assertEqual(3, len(self.darwin_exporter.h5.get_node("/Example")))

    def test_create_table_if_needed_replace_data(self):
        data = self.get_table_data()
        data["X"] *= 2
        self.darwin_exporter.create_table_if_needed("/", "Example", obj=data, drop_data=True)
        res = self.darwin_exporter.h5.get_node("/Example").read()
        numpy.testing.assert_equal(res, data)

    def test_create_table_if_needed_append_data(self):
        data = self.get_table_data()
        data["X"] *= 2
        expected = numpy.hstack((self.get_table_data(), data))
        self.darwin_exporter.create_table_if_needed("/", "Example", obj=data, dump_data=False)
        res = self.darwin_exporter.h5.get_node("/Example").read()
        numpy.testing.assert_equal(res, expected)


class HogConverterTest(unittest.TestCase):
    orthoxml_file = os.path.join(os.path.dirname(__file__), "hog-example.orthoXML")

    def setUp(self):
        self.h5 = tables.open_file("test", "w", driver="H5FD_CORE", driver_core_backing_store=0)
        self.h5.create_table(
            "/",
            "Entries",
            tablefmt.ProteinTable,
            obj=numpy.zeros(6, tables.dtype_from_descr(tablefmt.ProteinTable)),
        )

    def tearDown(self):
        self.h5.close()

    def test_extract_levels(self):
        conv = HogConverter(self.h5.root.Entries)
        levels = conv.convert_file(self.orthoxml_file)
        self.assertEqual(9, len(levels), "levels is broken: {}".format(levels))
        self.assertEqual(len(tables.dtype_from_descr(tablefmt.HOGsTable)), len(levels[0]))
        mammalia = next((x for x in levels if x[2] == "Mammalia"), None)
        self.assertAlmostEqual(1, mammalia[3], msg="CompletenessScore not what is expected")
        self.assertEqual(1, mammalia[4], "ImpliedLosses was not read from input xml")
        rodents = [x for x in levels if x[2] == "Rodents"]
        self.assertEqual(2, len(rodents), "expect 2 subhogs at level of Rodents")
        self.assertEqual([1, 0.5], [z[3] for z in rodents], "CompletenessScore does not match")

    def test_set_release_char(self):
        conv = HogConverter(self.h5.root.Entries, release_char="B")
        levels = conv.convert_file(self.orthoxml_file)
        self.assertTrue(all(map(lambda row: row[1].startswith("HOG:B0"), levels)))

    def test_invalid_release_char(self):
        for release_char in ("a", "AB", " "):
            with self.subTest(release_char=release_char):
                with self.assertRaises(ValueError):
                    conv = HogConverter(self.h5.root.Entries, release_char=release_char)


class ReadOnlyTestDbImporter(unittest.TestCase):
    importer = None
    h5_file_name = "TestDb.h5"

    @classmethod
    def setUpClass(cls) -> None:
        path = find_path_to_test_db(cls.h5_file_name)
        cls.importer = DarwinExporter(path, mode="read")

    def test_count_goterm_freqs(self):
        cnts, tot_per_ont = self.importer.collect_goterm_freqs()
        self.assertLess(sum(tot_per_ont), sum(cnts.values()))



import unittest
import numpy
from pyoma.browser.hoghelper import compare_levels


def merge_array_with_col(data):
    y = [tuple(x[0]) + (x[1],) for x in data]
    dt = data[0][0].dtype.descr + [("Event", "S10")]
    return numpy.array(y, dtype=dt)


class CompareHogLevels(unittest.TestCase):
    def setUp(self) -> None:
        self.parent = numpy.array(
            [
                (1, b"HOG:0000001", b"Parent"),
                (2, b"HOG:0000002", b"Parent"),
                (3, b"HOG:0000003", b"Parent"),
                (4, b"HOG:0000004", b"Parent"),
                (6, b"HOG:0000006.1a", b"Parent"),
                (6, b"HOG:0000006.1b", b"Parent"),
                (7, b"HOG:0000007.2c", b"Parent"),
            ],
            dtype=[("Fam", "i4"), ("ID", "S255"), ("Level", "S255")],
        )
        self.child = numpy.array(
            [
                (1, b"HOG:0000001", b"Child"),
                (2, b"HOG:0000002.1a", b"Child"),
                (4, b"HOG:0000004.1a", b"Child"),
                (4, b"HOG:0000004.1b", b"Child"),
                (5, b"HOG:0000005", b"Child"),
                (6, b"HOG:0000006.1a", b"Child"),
                (7, b"HOG:0000007.2c", b"Child"),
            ],
            dtype=self.parent.dtype,
        )

    def test_full_set_case(self):
        parent = self.parent
        child = self.child
        exp_res = [
            (child[0], "retained"),
            (child[1], "duplicated"),
            (parent[2], "lost"),
            (child[2], "duplicated"),
            (child[3], "duplicated"),
            (child[4], "gained"),
            (child[5], "retained"),
            (parent[5], "lost"),
            (child[6], "retained"),
        ]
        exp_res = merge_array_with_col(exp_res)
        exp_nr_dupl = 2
        diff, dupl = compare_levels(parent, child, return_duplication_events=True)
        self.assertEqual(exp_nr_dupl, dupl)
        self.assertTrue((exp_res == diff).all())

    def test_fam_1_to_5(self):
        child = self.child[self.child["Fam"] <= 5]
        parent = self.parent[self.parent["Fam"] <= 5]
        exp_res = [
            (child[0], "retained"),
            (child[1], "duplicated"),
            (parent[2], "lost"),
            (child[2], "duplicated"),
            (child[3], "duplicated"),
            (child[4], "gained"),
        ]
        exp_res = merge_array_with_col(exp_res)
        self.assertTrue((exp_res == compare_levels(parent, child)).all())

    def test_fam_2_to_6(self):
        child = self.child[(1 < self.child["Fam"]) & (self.child["Fam"] <= 6)]
        parent = self.parent[(1 < self.child["Fam"]) & (self.parent["Fam"] <= 6)]
        exp_res = [
            (child[0], "duplicated"),
            (parent[1], "lost"),
            (child[1], "duplicated"),
            (child[2], "duplicated"),
            (child[3], "gained"),
            (child[4], "retained"),
            (parent[4], "lost"),
        ]
        exp_res = merge_array_with_col(exp_res)
        self.assertTrue((exp_res == compare_levels(parent, child)).all())


class CompareLevelSpecialCases(unittest.TestCase):
    def test_with_lost_subfamilies(self):
        parent = numpy.array(
            [
                (1, b"HOG:0000001.1a", b"Parent"),
                (1, b"HOG:0000001.1b", b"Parent"),
                (1, b"HOG:0000001.1c", b"Parent"),
            ],
            dtype=[("Fam", "i4"), ("ID", "S255"), ("Level", "S255")],
        )
        child = numpy.array(
            [
                (1, b"HOG:0000001.1a", b"Child"),
                (1, b"HOG:0000001.1c.4b", b"Child"),
            ],
            dtype=[("Fam", "i4"), ("ID", "S255"), ("Level", "S255")],
        )
        exp_res = [
            (child[0], "retained"),
            (parent[1], "lost"),
            (child[1], "duplicated"),
        ]
        exp_res = merge_array_with_col(exp_res)

        exp_dups = 1
        diff, dupl = compare_levels(parent, child, return_duplication_events=True)
        self.assertTrue((exp_res == diff).all())
        self.assertEqual(exp_dups, dupl)


if __name__ == "__main__":
    unittest.main()



import functools
import queue
import signal
import threading
import uuid

import tables
import ete3
import multiprocessing as mp
import time
import tempfile
import pandas as pd
from datasketch import WeightedMinHashGenerator, MinHashLSHForest
from .pyhamutils import get_ham_treemap_from_row
from .hashutils import generate_treeweights, row2hash
from .tree_helper import (
    taxa_index,
    leaf_index,
    filter_tree,
    get_newick_tree_from_tax_db,
)
import logging
from logging.handlers import QueueHandler

logger = logging.getLogger(__name__)


class LSHBuilderBase(object):
    def __init__(self, db, numperm=256, taxfilter=None, taxmask=None, **kwargs):
        self.db = db
        tree_newick = get_newick_tree_from_tax_db(db.tax)
        full_tree = ete3.PhyloTree(tree_newick, format=1)
        self.tree = filter_tree(full_tree, root_node=taxmask, taxfilter=taxfilter)
        self.tree_newick = self.tree.write(format=1)
        self.taxa_index = taxa_index(self.tree)
        self.leaf_index = leaf_index(self.tree)
        self.tree_weights = generate_treeweights(self.tree, self.taxa_index)
        self.numperm = numperm
        self.wmg = WeightedMinHashGenerator(3 * len(self.taxa_index), sample_size=numperm, seed=1)


class BaseProfileBuilderProcess(mp.Process):
    def __init__(
        self,
        in_queue: mp.Queue,
        out_queue: mp.Queue,
        nr_workers_pre_step: mp.Value,
        log_queue: mp.Queue,
        control_queue: mp.Queue,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.in_queue = in_queue
        self.out_queue = out_queue
        self.log_queue = log_queue
        self.outstanding_dones = nr_workers_pre_step
        self.control_queue = control_queue
        self.quit_req = False
        self.proc_id = str(uuid.uuid4())

    def setup(self):
        pass

    def finalize(self):
        pass

    def _handle_signal(self, signum, frame):
        print("Stopping worker (pid: {}; name: {}): received signal {}".format(self.pid, self.name, signum))
        self.quit_req = True

    def run(self):
        self.setup()
        # Set signals for worker process
        signal.signal(signal.SIGINT, self._handle_signal)
        signal.signal(signal.SIGUSR2, self._handle_signal)
        signal.signal(signal.SIGTERM, self._handle_signal)

        nr_empty_cnt = 0
        while not self.quit_req and self.outstanding_dones.value > 0:
            try:
                item = self.in_queue.get(timeout=1)
                nr_empty_cnt = 0
            except queue.Empty:
                nr_empty_cnt += 1
                if nr_empty_cnt > 10:
                    logger.info(
                        "input queue seems to be empty ({}x), but expecting still some chunks".format(nr_empty_cnt)
                    )
                self.control_queue.put((self.proc_id, "NO WORK"))
                continue

            if item is None:
                logger.info("received sentinel from previous pipeline step")
                with self.outstanding_dones.get_lock():
                    self.outstanding_dones.value -= 1
                logger.info("outstanding_dones: %s", self.outstanding_dones)
                self.control_queue.put((self.proc_id, "DONE"))
                logger.debug("sent DONE info to control queue (%s)", self.proc_id)
            else:
                result = self.handle_input(item)
                if result is not None:
                    self.out_queue.put(result)
                self.control_queue.put((self.proc_id, "JOB_HANDELED"))
        if not self.quit_req:
            self.out_queue.put(None)  # signal end of work for this worker
            logger.info("sent sentinel to next stage")
        else:
            logger.info("received termination signal")
        self.finalize()

    def handle_input(self, item):
        raise NotImplementedError("This method must be overwritten")


class SourceProcess(BaseProfileBuilderProcess):
    def __init__(self, out_queue, **kwargs):
        super().__init__(in_queue=None, out_queue=out_queue, nr_workers_pre_step=None, **kwargs)

    def _put_in_queue_with_status(self, input):
        while True:
            try:
                self.out_queue.put(input, timeout=5)
                break
            except queue.Full:
                qsize = "?"
                try:
                    qsize = self.out_queue.qsize()
                except NotImplementedError:
                    pass
                logger.info("Queue full. (approx size: {}) Cannot put item {} into output queue".format(qsize, input))
            self.control_queue.put((self.proc_id, "ALIVE"))

    def run(self):
        self.setup()
        for input in self.generate_data():
            self._put_in_queue_with_status(input)
        self.out_queue.put(None)
        self.control_queue.put((self.proc_id, "DONE"))
        logger.info("sent sentinel to next stage")
        self.finalize()

    def generate_data(self):
        raise NotImplementedError("SourceProcess should implement this method")


class ProfileBuilder(BaseProfileBuilderProcess):
    def __init__(self, db_path, **kwargs):
        super().__init__(**kwargs)
        self.db_path = db_path
        self.builder = None
        self.ham_pipeline = None
        self.hash_pipeline = None

    def setup(self):
        from ..db import Database

        self.builder = LSHBuilderBase(db=Database(self.db_path))
        self.ham_pipeline = functools.partial(get_ham_treemap_from_row, tree=self.builder.tree_newick)
        self.hash_pipeline = functools.partial(
            row2hash,
            taxa_index=self.builder.taxa_index,
            species_index=self.builder.leaf_index,
            treeweights=self.builder.tree_weights,
            wmg=self.builder.wmg,
        )
        print("Worker {} init".format(self.name))

    def finalize(self):
        self.builder.db.close()

    def handle_input(self, df):
        print("handling df: {}".format(df))
        df["tree"] = df[["Fam", "ortho"]].apply(self.ham_pipeline, axis=1)
        df[["hash", "rows", "species"]] = df[["Fam", "tree"]].apply(self.hash_pipeline, axis=1)
        return df[["Fam", "hash", "species"]]


class Collector(BaseProfileBuilderProcess):
    def __init__(self, db_path, tmp_file, **kwargs):
        super().__init__(**kwargs)
        self.tmp_file = tmp_file
        self.db_path = db_path

    def _init_tmp_h5(self, root):
        print(root)
        hashes = self.h5.create_carray(
            root,
            "hashes",
            createparents=True,
            atom=tables.Int64Atom(),
            shape=(
                self.builder.db.get_nr_toplevel_hogs() + 1,
                self.builder.numperm * 2,
            ),
            filters=tables.Filters(complevel=3, complib="blosc"),
        )
        species = self.h5.create_carray(
            root,
            "species_profile",
            createparents=True,
            atom=tables.Int8Atom(),
            shape=(
                self.builder.db.get_nr_toplevel_hogs() + 1,
                len(self.builder.leaf_index),
            ),
            filters=tables.Filters(complevel=3, complib="blosc"),
        )
        lsh_forest = self.h5.create_vlarray(
            root,
            "min_hash_lsh_forest",
            createparents=True,
            atom=tables.ObjectAtom(),
            filters=tables.Filters(complevel=3, complib="blosc"),
        )
        lsh_tree = self.h5.create_vlarray(
            root,
            "species_tree",
            createparents=True,
            atom=tables.ObjectAtom(),
            filters=tables.Filters(complevel=3, complib="blosc"),
        )
        lsh_tree.append(self.builder.tree)
        self.h5.set_node_attr(root, "num_perm", self.builder.numperm)
        return hashes, species, lsh_forest

    def setup(self):
        from ..db import Database

        self.builder = LSHBuilderBase(db=Database(self.db_path))
        self.save_start = time.time()
        self.global_time = time.time()
        self.count = 0
        self.forest = MinHashLSHForest(num_perm=self.builder.numperm)
        self.h5 = tables.open_file(self.tmp_file, "w")
        root = "/HOGProfile/ALL"
        # if self.builder.tree.name != "0":
        #     root = "/HOGProfile/{}".format(self.builder.tree.name)
        print("storing results in {}:{}".format(self.tmp_file, root))

        self.hashes_matrix, self.species_matrix, self.forest_arr = self._init_tmp_h5(root)
        print("Collector process initialized")

    def handle_input(self, df: pd.DataFrame):
        print("handling hash df: {}".format(df))
        if not df.empty:
            hashes = df["hash"].to_dict()
            hashes = {fam: hashes[fam] for fam in hashes if hashes[fam]}
            for fam, hashvals in hashes.items():
                self.forest.add(str(fam), hashvals)
                self.hashes_matrix[fam, :] = hashvals.hashvalues.ravel()
                self.species_matrix[fam, :] = df["species"][fam]
                self.count += 1
            if time.time() - self.save_start > 200:
                print("Saving current results: {}".format(time.time() - self.global_time))
                self.forest.index()
                print(
                    "Checking consistency of search index. Fam {} --> {}".format(
                        fam, self.forest.query(hashes[fam], k=10)
                    )
                )
                self.save_start = time.time()
                print("saved results")
        else:
            print("Empty dataframe in Collector: {}".format(df))

    def finalize(self):
        print("received all results. wrapping up...")
        print("computed minhashes: {}".format(self.count))
        self.forest.index()
        self.forest_arr.append(self.forest)
        self.h5.flush()
        self.h5.close()
        self.builder.db.close()
        print("stored all data in temporary file")


class HogGenerator(SourceProcess):
    def __init__(self, db_path, max_hogsize=None, min_hogsize=100, chunk_size=100, **kwargs):
        super().__init__(**kwargs)
        self.db_path = db_path
        self.max_hogsize = max_hogsize
        self.min_hogsize = min_hogsize
        self.chunk_size = chunk_size

    def generate_data(self):
        from ..db import Database

        db = Database(self.db_path)
        nr_groups = db.get_nr_toplevel_hogs()
        chunk = {}
        logger.debug("nr of toplevel hogs: %d", nr_groups)
        for fam in range(1, nr_groups + 1):
            try:
                orthoxml = db.get_orthoxml(fam).decode()
            except ValueError as e:
                continue
            nr_species = orthoxml.count("<species name=")
            if (self.max_hogsize is None or nr_species < self.max_hogsize) and (
                self.min_hogsize is None or nr_species >= self.min_hogsize
            ):
                chunk[fam] = {"ortho": orthoxml}
            if len(chunk) > self.chunk_size:
                df = pd.DataFrame.from_dict(chunk, orient="index")
                df["Fam"] = df.index
                print("yielding chunk")
                yield df
                chunk = {}
        # send last chunk
        if len(chunk) > 0:
            df = pd.DataFrame.from_dict(chunk, orient="index")
            df["Fam"] = df.index
            yield df
        print("all hogs generated")
        db.close()


def rootlogger_configurer(queue):
    h = QueueHandler(queue)
    root = logging.getLogger()
    root.addHandler(h)
    root.setLevel(logging.DEBUG)


def logger_listener(log_queue):
    root = logging.getLogger()
    console_handler = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s %(processName)-10s %(name)s %(levelname)-8s %(message)s")
    console_handler.setFormatter(formatter)
    root.addHandler(console_handler)
    root.setLevel(logging.DEBUG)
    for record in iter(log_queue.get, None):
        logger = logging.getLogger(record.name)
        logger.handle(record)


class PipelineControllerThread(threading.Thread):
    def __init__(self, control_queue, log_queue, processes, time_until_kill=1800, **kwargs):
        super().__init__(**kwargs)
        self.control_queue = control_queue
        self.log_queue = log_queue
        self.processes = {p.proc_id: [0, p] for p in processes}
        self.time_until_kill = time_until_kill
        self._last_timeout_check = time.time()

    def _process_alive_info(self, item):
        proc_id, flag = item
        logger.debug(item)
        if flag == "DONE":
            self.processes.pop(proc_id)
        else:
            self.processes[proc_id][0] = time.time()

    def _kill_orphan_procs(self):
        to_rem = []
        for proc_id, proc_info in self.processes.items():
            last_seen, proc = proc_info[0], proc_info[1]
            if last_seen == 0:
                logger.info("process %s[%s] not yet started?", proc.name, proc.proc_id)
            elif time.time() - last_seen > self.time_until_kill:
                logger.warning(
                    "process %s[%s] seems to be dead. No alive signal since %fsec.",
                    proc.name,
                    proc.proc_id,
                    time.time() - last_seen,
                )
                proc.terminate()
                proc.out_queue.put(None)
                to_rem.append(proc_id)
        for proc_id in to_rem:
            self.processes.pop(proc_id)
        logger.info("killed %d orphan processes", len(to_rem))

    def run(self):
        logger.info("starting controler thread")
        nr_empty_cnt = 0
        while len(self.processes) > 0:
            try:
                item = self.control_queue.get(timeout=30)
                nr_empty_cnt = 0
                if item is None:
                    break  # None indicates end of pipeline
                self._process_alive_info(item)
            except queue.Empty:
                nr_empty_cnt += 1
                if nr_empty_cnt > 10:
                    logger.info(
                        "control queue seems to be empty ({}x), but expecting still some running jobs".format(
                            nr_empty_cnt
                        )
                    )
            if time.time() - self._last_timeout_check > 10:
                self._kill_orphan_procs()
                self._last_timeout_check = time.time()


class Stage(object):
    def __init__(self, process, nr_procs, **kwargs):
        self.process = process
        self.nr_procs = nr_procs
        self.kwargs = kwargs
        self.outstanding_dones = None
        self.in_queue = None
        self.out_queue = None


class Pipeline(object):
    def __init__(self):
        self.stages = []

    def add_stage(self, stage: Stage):
        self.stages.append(stage)

    def run(self):
        log_queue = mp.Queue()
        logger_process = mp.Process(target=logger_listener, args=(log_queue,))
        logger_process.start()
        rootlogger_configurer(log_queue)
        print("setting up pipeline")
        control_queue = mp.Queue()

        procs = []
        for i in range(len(self.stages) - 1, 0, -1):
            stage = self.stages[i]
            stage.outstanding_dones = mp.Value("d", self.stages[i - 1].nr_procs)
            stage.in_queue = mp.Queue(maxsize=10 * mp.cpu_count())
            stage.out_queue = self.stages[i + 1].in_queue if i < len(self.stages) - 1 else mp.Queue()
            for k in range(stage.nr_procs):
                p = stage.process(
                    in_queue=stage.in_queue,
                    out_queue=stage.out_queue,
                    nr_workers_pre_step=stage.outstanding_dones,
                    log_queue=log_queue,
                    control_queue=control_queue,
                    daemon=True,
                    **stage.kwargs,
                )
                procs.append(p)
                p.start()
        # add source process
        stage = self.stages[0]
        stage.out_queue = self.stages[1].in_queue
        for k in range(stage.nr_procs):
            p = stage.process(
                out_queue=stage.out_queue,
                log_queue=log_queue,
                control_queue=control_queue,
                **stage.kwargs,
            )
            procs.append(p)
            p.start()
        control_thread = PipelineControllerThread(control_queue=control_queue, log_queue=log_queue, processes=procs)
        control_thread.start()
        print("all processes started")
        try:
            for p in procs:
                p.join()
        except KeyboardInterrupt:
            print("keyboard interupt in main loop")
            time.sleep(20)
        log_queue.put(None)
        control_queue.put(None)
        logger_process.join()
        control_thread.join()

        print("successfully joined all processes")


def compute_profiles(db_path, min_hogsize=100, max_hogsize=None, nr_procs=None):
    pipeline = Pipeline()
    with tempfile.NamedTemporaryFile(suffix=".h5", delete=False) as h5_tmp:
        tmp_file = h5_tmp.name
    if nr_procs is None:
        nr_procs = mp.cpu_count()

    pipeline.add_stage(
        Stage(
            HogGenerator,
            nr_procs=1,
            db_path=db_path,
            min_hogsize=min_hogsize,
            max_hogsize=max_hogsize,
        )
    )
    pipeline.add_stage(Stage(ProfileBuilder, nr_procs=nr_procs, db_path=db_path))
    pipeline.add_stage(Stage(Collector, nr_procs=1, db_path=db_path, tmp_file=tmp_file))
    print("generated pipeline. about to starting it")
    pipeline.run()
    print("finished computing profiles")

    with tables.open_file(db_path, "a") as db, tables.open_file(tmp_file, "r") as tmp:
        tmp.root._f_copy_children(db.root, recursive=True, overwrite=True)
    print("Finished writing everything")



from __future__ import division, print_function, unicode_literals


class PyOmaException(Exception):
    pass


class IdException(PyOmaException):
    pass


class InvalidTaxonId(IdException):
    pass


class DBVersionError(PyOmaException):
    pass


class DBConsistencyError(PyOmaException):
    pass


class DBOutdatedError(PyOmaException):
    pass


class InvalidId(IdException):
    pass


class InvalidOmaId(InvalidId):
    pass


class UnknownIdType(IdException):
    pass


class UnknownSpecies(IdException):
    pass


class OutdatedHogId(InvalidId):
    def __init__(self, hog_id):
        super().__init__("Outdated HOG ID: {}".format(hog_id), hog_id)
        self.outdated_hog_id = hog_id


class Singleton(Exception):
    def __init__(self, entry, msg=None):
        super(Singleton, self).__init__(msg)
        self.entry = entry


class NoReprEntry(Exception):
    pass


class AmbiguousID(Exception):
    def __init__(self, message, candidates):
        super(AmbiguousID, self).__init__(message, candidates)
        self.candidates = candidates


class TooUnspecificQuery(PyOmaException):
    def __init__(self, query, hits):
        super().__init__("Too unspecific query '{}' results in {} hits".format(query, hits))
        self.query = query
        self.hits = hits



import numpy
import tables

"""This module contains the definitions of the database tables
used in the browser database. Some of these tables are used
multiple times, e.g. the PairwiseRelationTable is used
for each genome pair.

From these table definitions one can easily extract the numpy
dtype that can hold the data:

   >>>tables.dtype_from_descr(HOGsTable)
   dtype([('Fam', '<i4'), ('ID', 'S255'), ('Level', 'S255')])
"""


class HOGsTable(tables.IsDescription):
    Fam = tables.Int32Col(pos=1)
    ID = tables.StringCol(255, pos=2)
    Level = tables.StringCol(255, pos=3)
    CompletenessScore = tables.Float32Col(pos=4, dflt=-1)
    ImpliedLosses = tables.Int32Col(pos=5, dflt=-1)
    NrMemberGenes = tables.Int32Col(pos=6, dflt=-1)
    IsRoot = tables.BoolCol(pos=7, dflt=False)
    IdxPerLevelTable = tables.Int32Col(pos=8, dflt=-1)


class OrthoXmlHogTable(tables.IsDescription):
    Fam = tables.UInt32Col(pos=0)
    HogBufferOffset = tables.Int64Col(pos=1)
    HogBufferLength = tables.UInt32Col(pos=2)
    HogAugmentedBufferOffset = tables.Int64Col(pos=3)
    HogAugmentedBufferLength = tables.UInt32Col(pos=4)


class AbstractSyntenyRels(tables.IsDescription):
    Weight = tables.Float16Col(pos=2)
    Evidence = tables.EnumCol(
        tables.Enum({"linearized": 1, "parsimonious": 2, "any": 4}),
        pos=3,
        dflt="any",
        base="uint8",
    )
    LCA_taxid = tables.Int32Col(pos=4, dflt=-1)


class AncestralSyntenyRels(AbstractSyntenyRels):
    HogRow1 = tables.UInt32Col(pos=0)
    HogRow2 = tables.UInt32Col(pos=1)


class ExtantSyntenyRels(AbstractSyntenyRels):
    EntryNr1 = tables.UInt32Col(pos=0)
    EntryNr2 = tables.UInt32Col(pos=1)


class ProteinTable(tables.IsDescription):
    EntryNr = tables.UInt32Col(pos=1)
    SeqBufferOffset = tables.UInt64Col(pos=2)
    SeqBufferLength = tables.UInt32Col(pos=3)
    OmaGroup = tables.UInt32Col(pos=4, dflt=0)
    OmaHOG = tables.StringCol(255, pos=5, dflt=b"")
    Chromosome = tables.StringCol(255, pos=6)
    LocusStart = tables.UInt32Col(pos=7)
    LocusEnd = tables.UInt32Col(pos=8)
    LocusStrand = tables.Int8Col(pos=9, dflt=1)
    AltSpliceVariant = tables.Int32Col(pos=10, dflt=0)
    CanonicalId = tables.StringCol(20, pos=11, dflt=b"")
    CDNABufferOffset = tables.UInt64Col(pos=12)
    CDNABufferLength = tables.UInt32Col(pos=13)
    MD5ProteinHash = tables.StringCol(32, pos=14)
    DescriptionOffset = tables.UInt32Col(pos=15)
    DescriptionLength = tables.UInt16Col(pos=16)
    SubGenome = tables.StringCol(1, pos=17, dflt=b"")


class ProteinCacheInfo(tables.IsDescription):
    EntryNr = tables.UInt32Col(pos=0)
    NrPairwiseOrthologs = tables.UInt32Col(pos=1)
    NrHogInducedPWOrthologs = tables.UInt32Col(pos=2)
    NrHogInducedPWParalogs = tables.UInt32Col(pos=3)
    NrOMAGroupOrthologs = tables.UInt32Col(pos=4)
    NrAnyOrthologs = tables.UInt32Col(pos=5)


class LocusTable(tables.IsDescription):
    EntryNr = tables.UInt32Col(pos=1)
    Start = tables.UInt32Col(pos=2)
    End = tables.UInt32Col(pos=3)
    Strand = tables.Int8Col(pos=4)


class PairwiseRelationTable(tables.IsDescription):
    EntryNr1 = tables.UInt32Col(pos=0)
    EntryNr2 = tables.UInt32Col(pos=1)
    RelType = tables.EnumCol(
        tables.Enum(
            {
                "1:1": 0,
                "1:n": 1,
                "m:1": 2,
                "m:n": 3,
                "close paralog": 4,
                "homeolog": 5,
                "n/a": 6,
            }
        ),
        "n/a",
        base="uint8",
        pos=2,
    )
    Score = tables.Float32Col(pos=3, dflt=-1)
    Distance = tables.Float32Col(pos=4, dflt=-1)
    AlignmentOverlap = tables.Float16Col(pos=5, dflt=-1)
    SyntenyConservationLocal = tables.Float16Col(pos=6, dflt=-1)
    Confidence = tables.Float16Col(pos=7, dflt=-1)


class XRefTable(tables.IsDescription):
    EntryNr = tables.UInt32Col(pos=1)
    XRefSource = tables.EnumCol(
        tables.Enum(
            {
                "UniProtKB/SwissProt": 0,
                "UniProtKB/TrEMBL": 10,
                "Ensembl Protein": 20,
                "Ensembl Gene": 25,
                "Ensembl Transcript": 30,
                "RefSeq": 40,
                "EntrezGene": 50,
                "FlyBase": 60,
                "WormBase": 65,
                "EnsemblGenomes": 70,
                "NCBI": 75,
                "EMBL": 80,
                "SourceID": 95,
                "SourceAC": 100,
                "HGNC": 105,
                "Gene Name": 110,
                "Synonym": 115,
                "Protein Name": 120,
                "ORF Name": 125,
                "Ordered Locus Name": 130,
                "PDB": 148,
                "Swiss Model": 150,
                "STRING": 151,
                "neXtProt": 152,
                "Bgee": 153,
                "EPD": 154,
                "ChEMBL": 156,
                "GlyConnect": 157,
                "SwissPalm": 158,
                "DisGeNET": 159,
                "WikiGene": 160,
                "IPI": 240,
                "GI": 241,
                "n/a": 255,
            }
        ),  # last line: deprecated systems
        "n/a",
        base="uint8",
        pos=2,
    )
    XRefId = tables.StringCol(50, pos=3)
    Verification = tables.EnumCol(
        tables.Enum({"exact": 0, "unchecked": 2, "modified": 4}),
        "unchecked",
        base="uint8",
        pos=4,
    )


class GeneOntologyTable(tables.IsDescription):
    EntryNr = tables.UInt32Col(pos=1)
    TermNr = tables.UInt32Col(pos=2)
    Evidence = tables.StringCol(3, pos=3)
    Reference = tables.StringCol(255, pos=4)


class AncestralGeneOntologyTable(tables.IsDescription):
    HogRow = tables.UInt32Col(pos=1)
    TermNr = tables.UInt32Col(pos=2)
    Score = tables.UInt8Col(pos=3)
    RawScore = tables.Float16Col(pos=4)


class GeneOntologyTermCounts(tables.IsDescription):
    TermNr = tables.UInt32Col(pos=1)
    Counts = tables.UInt32Col(pos=2)


class ECTable(tables.IsDescription):
    EntryNr = tables.UInt32Col(pos=1)
    ECacc = tables.StringCol(16, pos=2)


class GenomeTable(tables.IsDescription):
    NCBITaxonId = tables.Int32Col(pos=0)
    UniProtSpeciesCode = tables.StringCol(5, pos=1)
    TotEntries = tables.UInt32Col(pos=2)
    TotAA = tables.UInt32Col(pos=3)
    EntryOff = tables.UInt32Col(pos=4)
    SciName = tables.StringCol(255, pos=5)
    CommonName = tables.StringCol(64, pos=6)
    SynName = tables.StringCol(64, pos=7)
    Release = tables.StringCol(128, pos=8)
    Url = tables.StringCol(255, pos=9)
    Source = tables.StringCol(255, pos=10)
    Date = tables.Time32Col(pos=11)
    IsPolyploid = tables.BoolCol(pos=12)
    TotGenes = tables.UInt32Col(pos=13, dflt=0)


class TaxonomyTable(tables.IsDescription):
    NCBITaxonId = tables.Int32Col(pos=0)
    ParentTaxonId = tables.Int32Col(pos=1)
    Name = tables.StringCol(255, pos=2)
    Age = tables.Float16Col(pos=3, dflt=numpy.nan)


class DomainTable(tables.IsDescription):
    EntryNr = tables.UInt32Col(pos=0)
    DomainId = tables.StringCol(20, pos=1)
    Coords = tables.StringCol(255, pos=2)


class DomainDescriptionTable(tables.IsDescription):
    DomainId = tables.StringCol(20, pos=0)
    Source = tables.StringCol(11, pos=1)
    Description = tables.StringCol(150, pos=2)


class HOGDomainArchPrevalenceTable(tables.IsDescription):
    Fam = tables.UInt32Col(pos=0)
    ReprEntryNr = tables.UInt32Col(pos=1)
    ReprEntryLength = tables.UInt32Col(pos=2)
    TopLevel = tables.StringCol(255, pos=3)
    FamSize = tables.UInt32Col(pos=4)
    PrevCount = tables.UInt32Col(pos=5)


class HOGDomainPresenceTable(tables.IsDescription):
    Offset = tables.UInt32Col(pos=0)
    DomainId = tables.StringCol(20, pos=1)


class GroupsizeHistogram(tables.IsDescription):
    Size = tables.UInt32Col(pos=0)
    Count = tables.UInt32Col(pos=1)


class OmaGroupTable(tables.IsDescription):
    GroupNr = tables.UInt32Col(pos=0)
    Fingerprint = tables.StringCol(7, pos=1)
    KeywordOffset = tables.UInt32Col(pos=2)
    KeywordLength = tables.UInt16Col(pos=3)
    NrMembers = tables.UInt16Col(pos=4)


class RootHOGMetaTable(tables.IsDescription):
    FamNr = tables.UInt32Col(pos=0)
    KeywordOffset = tables.UInt32Col(pos=1)
    KeywordLength = tables.UInt32Col(pos=2)
    FamDataJsonOffset = tables.UInt64Col(pos=3)
    FamDataJsonLength = tables.UInt32Col(pos=4)



import collections
import itertools
import logging
import multiprocessing
import os
import re
import time
from functools import partial
from tqdm import tqdm

import numpy
import tables
from datasketch import MinHash, MinHashLSH, LeanMinHash

from .db import Database
from .hogprofile.build import BaseProfileBuilderProcess, SourceProcess, Pipeline, Stage

logger = logging.getLogger(__name__)
MinHash256 = partial(MinHash, seed=1, num_perm=256)
LeanMinHash256 = partial(LeanMinHash, seed=1)


class HogHasher(object):
    def __init__(self, db: Database):
        self.db = db
        self.xrefs = db.id_mapper["XRef"]

    def analyze_fam(self, fam_nr):
        members = self.db.member_of_fam(fam_nr)
        minhashes = collections.defaultdict(MinHash256)
        t_start = t0 = time.time()
        for i, e in enumerate(members):
            hog_id = e["OmaHOG"]
            prot_id = e["CanonicalId"]
            if len(prot_id) == e.dtype["CanonicalId"].itemsize:
                # probably overflow, get full id from xref
                for ref in self.xrefs.iter_xrefs_for_entry_nr(e["EntryNr"]):
                    if ref["xref"].encode("utf-8").startswith(prot_id):
                        prot_id = ref["xref"].encode("utf-8")
                        break
            for p in re.finditer(rb"\.", hog_id):
                minhashes[hog_id[: p.start()]].update(prot_id)
            minhashes[hog_id].update(prot_id)
            if time.time() - t0 > 10:
                logger.debug(
                    f"working since {time.time()-t_start}sec on fam {fam_nr}. Done {i} out of {len(members)} proteins"
                )
                t0 = time.time()
        return minhashes


class LSHBuilder(object):
    def __init__(self, hash_file, mode="r", threshold=0.7):
        if mode not in ("r", "a", "w"):
            raise ValueError("invalid mode string ``%s``. Allowed modes are: " "'r', 'a' and 'w'" % mode)
        if mode == "r":
            if not os.path.exists(hash_file):
                raise IOError('file "{}" does not exist'.format(hash_file))
            self.h5, self.lsh, self.hogid2row = self._load_hash_file(hash_file, mode)
        elif mode == "a" and os.path.exists(hash_file):
            self.h5, self.lsh, self.hogid2row = self._load_hash_file(hash_file, mode)
        elif mode == "w" or mode == "a" and not os.path.exists(hash_file):
            self.lsh = MinHashLSH(threshold=threshold, num_perm=256)
            self.hogid2row = {}
            self.h5 = self.init_hash_table_file(hash_file)
        self.hashes = self.h5.get_node("/hashes")
        self.hogids = self.h5.get_node("/hogids")
        self.threshold = threshold
        self._readonly = mode == "r"

    def _open_hdf5(self, filename, mode="w"):
        filters = None
        if mode == "w":
            filters = tables.Filters(complevel=1, complib="blosc", shuffle=True)
        return tables.open_file(filename, mode=mode, filters=filters)

    def init_hash_table_file(self, hash_file):
        h5 = self._open_hdf5(hash_file, mode="w")
        h5.create_earray("/", "hashes", atom=tables.Int64Atom(), shape=(0, 256), expectedrows=1e6)
        h5.create_earray(
            "/",
            "hogids",
            atom=tables.StringAtom(itemsize=255),
            shape=(0,),
            expectedrows=1e6,
        )
        h5.create_vlarray("/", "lsh_obj", atom=tables.ObjectAtom())
        return h5

    def _load_hash_file(self, hash_file, mode="r"):
        h5 = self._open_hdf5(hash_file, mode=mode)
        lsh_obj_arr = h5.get_node("/lsh_obj")
        lsh = lsh_obj_arr[-2]
        hogid2row = lsh_obj_arr[-1]
        return h5, lsh, hogid2row

    def close(self):
        if self.h5.mode != "r":
            lsh_obj_arr = self.h5.get_node("/lsh_obj")
            lsh_obj_arr.append(self.lsh)
            lsh_obj_arr.append(self.hogid2row)
            self.hashes.flush()
            self.hogids.flush()
        self.h5.close()

    def add_minhashes(self, it):
        for hogid, minhash in it:
            self.hashes.append([minhash.digest()])
            self.hogids.append([hogid])
        self.hashes.flush()
        self.hogids.flush()

    def compute_lsh(self):
        lsh = MinHashLSH(threshold=self.threshold, num_perm=256)
        hog2row = {}
        for row, (hogid, hashvals) in enumerate(itertools.zip_longest(self.hogids, self.hashes)):
            hog2row[hogid] = row
            lsh.insert(hogid, LeanMinHash256(hashvalues=hashvals))
        self.lsh = lsh
        self.hogid2row = hog2row

    def query(self, key, minhash):
        """query with a minhash, returns a list of tuples of
        (query-key, target-key, jaccard)"""
        candidates = self.lsh.query(minhash)
        for c in candidates:
            hashvals = self.hashes[self.hogid2row[c]]
            h = MinHash256(hashvalues=hashvals)
            yield key, c, minhash.jaccard(h)


class FamGenerator(SourceProcess):
    def __init__(self, fam_generator=None, **kwargs):
        super().__init__(**kwargs)
        self.fams_to_process = fam_generator

    def generate_data(self):
        for fam in self.fams_to_process:
            yield fam
        logger.info("all families put to queue")


class HashWorker(BaseProfileBuilderProcess):
    def __init__(self, db_path, **kwargs):
        super().__init__(**kwargs)
        self.db_path = db_path
        self.db = None
        self.hasher = None
        self.handled_queries = None

    def setup(self):
        self.db = Database(self.db_path)
        self.hasher = HogHasher(self.db)
        self.handled_queries = 0

    def handle_input(self, fam):
        logger.info("computing hashes for family %s", fam)
        if self.handled_queries > 10000:
            self.db.close()
            logger.info("resetting database handle")
            self.setup()
        logger.info("start chewing on fam %s...", fam)
        t0 = time.time()
        hashes = self.hasher.analyze_fam(fam)
        logger.info("... done with fam %s. Took %f sec", fam, time.time() - t0)
        self.handled_queries += 1
        return hashes

    def finalize(self):
        self.db.close()


class Collector(BaseProfileBuilderProcess):
    def __init__(self, output_path, **kwargs):
        super().__init__(**kwargs)
        self.output_path = output_path

    def setup(self):
        self.lsh_builder = LSHBuilder(self.output_path, mode="a")

    def handle_input(self, hashes):
        logger.info("build lsh for {} hashes ({})".format(len(hashes), next(iter(hashes))))
        self.lsh_builder.add_minhashes(hashes.items())

    def finalize(self):
        self.lsh_builder.close()


def generator_of_unprocessed_fams(db_path, lsh_path=None):
    def _get_nr_families(db_path):
        db = Database(db_path)
        nr_hogs = db.get_nr_toplevel_hogs()
        db.close()
        logger.info("Found %d families to process", nr_hogs)
        return nr_hogs

    def _load_unprocessed_fams(db_path, lsh_path):
        db = Database(db_path)
        with tables.open_file(lsh_path, "r") as lsh_h5:
            processed_fams = set(db.parse_hog_id(x) for x in lsh_h5.get_node("/hogids"))
        remaining = set(range(1, db.get_nr_toplevel_hogs() + 1)) - processed_fams
        db.close()
        logger.info("Found %d unprocessed families", len(remaining))
        return remaining

    if lsh_path is None or not os.path.exists(lsh_path):
        fams_to_process = range(1, _get_nr_families(db_path) + 1)
    else:
        fams_to_process = _load_unprocessed_fams(db_path, lsh_path)
    return fams_to_process


def compute_minhashes_for_db(db_path, output_path, nr_procs=None):
    if nr_procs is None:
        nr_procs = multiprocessing.cpu_count()

    fams_to_process = generator_of_unprocessed_fams(db_path, output_path)
    while True:
        pipeline = Pipeline()
        pipeline.add_stage(Stage(FamGenerator, nr_procs=1, fam_generator=fams_to_process))
        pipeline.add_stage(Stage(HashWorker, nr_procs=nr_procs, db_path=db_path))
        pipeline.add_stage(Stage(Collector, nr_procs=1, output_path=output_path))
        print("setup pipeline, about to start it.")
        pipeline.run()
        print("finished with computing the MinHashLSH for {}".format(db_path))

        fams_to_process = set(generator_of_unprocessed_fams(db_path, output_path))
        if len(fams_to_process) == 0:
            break
    print("for sure all families processes. we're done!")


def compare_versions(output_file, target_path, *old_path):
    lsh = LSHBuilder(target_path, mode="r")
    lsh.compute_lsh()
    with tables.open_file(
        output_file,
        "w",
        filters=tables.Filters(complevel=1, complib="blosc", shuffle=True),
    ) as h5_map:
        tab = h5_map.create_table(
            "/",
            "hogmap",
            description=numpy.dtype([("Old", "S255"), ("New", "S255"), ("Jaccard", "f4")]),
        )
        dubious = h5_map.create_earray(
            "/",
            "dubious",
            atom=tables.StringAtom(itemsize=255),
            shape=(0,),
            expectedrows=1e5,
        )
        for old in old_path:
            old = LSHBuilder(old, mode="r")
            for old_id, old_hashvals in tqdm(itertools.zip_longest(old.hogids, old.hashes), total=len(old.hogids)):
                minhash = LeanMinHash256(hashvalues=old_hashvals)
                candidates = sorted(lsh.query(old_id, minhash), key=lambda x: -x[2])
                logger.debug("old_id: %s: candidates: %s", old_id, candidates)
                if len(candidates) > 10 and candidates[6][2] > 0.9:
                    # this is a dubious node, store it for now, maybe try to recover later.
                    # it seems that this happens if the CanonicalId is truncated and hence maps to several ids.
                    dubious.append([old_id])
                    continue
                if len(candidates) > 0 and candidates[0][2] > 0.6:
                    tab.append([(old_id, candidates[0][1], candidates[0][2])])
                    if candidates[0][2] < 1:
                        other_cands = [(old_id, c[1], c[2]) for c in candidates[1:]]
                        if len(other_cands) > 0:
                            tab.append(other_cands)
            tab.flush()
            dubious.flush()
            old.close()
        # build index of Old ids
        tab.colinstances["Old"].create_csindex()


def build_lookup(target_db, old_dbs, nr_procs=None):
    def lsh_fn_from_db_path(dbpath, modif=None):
        dbname = os.path.splitext(os.path.basename(dbpath))[0]
        modif_str = "-{}".format(modif) if modif is not None else ""
        lsh_fn = "{}{}.hog-lsh.h5".format(dbname, modif_str)
        return lsh_fn

    target_lsh_fn = lsh_fn_from_db_path(target_db)
    compute_minhashes_for_db(target_db, target_lsh_fn, nr_procs=nr_procs)

    old_lsh_paths = []
    for k, dbpath in enumerate(old_dbs):
        cached_lsh_fn = dbpath.replace(".h5", ".hog-lsh.h5")
        if not os.path.exists(cached_lsh_fn):
            cached_lsh_fn = lsh_fn_from_db_path(dbpath, modif=k)
            print("computing lsh for {} - storing in {}".format(dbpath, cached_lsh_fn))
            compute_minhashes_for_db(dbpath, cached_lsh_fn, nr_procs=nr_procs)
        old_lsh_paths.append(cached_lsh_fn)

    hogmap_name = os.path.splitext(os.path.basename(target_db))[0] + ".hogmap.h5"
    compare_versions(hogmap_name, target_lsh_fn, old_lsh_paths)



import collections
import io
import multiprocessing
import itertools
import re
import numpy
import pandas
import tables
import logging
import os
from pathlib import Path
from typing import List, Tuple, Set, Union
from .hogprofile.build import Pipeline, SourceProcess, BaseProfileBuilderProcess, Stage

logger = logging.getLogger(__name__)


class GeneEntries:
    def __init__(self, enrs, main):
        self.main = main
        self.enrs = enrs

    def entrynr_slices(self):
        if self.enrs[-1] - self.enrs[0] == len(self.enrs) - 1:
            return (slice(self.enrs[0], self.enrs[-1] + 1, 1),)
        slices, i0 = [], 0
        for i in range(len(self.enrs)):
            if self.enrs[i] - self.enrs[i0] == i - i0:
                continue
            slices.append(slice(self.enrs[i0], self.enrs[i - 1] + 1, 1))
            i0 = i
        slices.append(slice(self.enrs[i0], self.enrs[i] + 1, 1))
        return tuple(slices)


class SpliceVariantHelper:
    def __init__(self, h5, alt_array=None):
        if alt_array is not None:
            lookup = numpy.copy(alt_array)
            lookup[numpy.where(lookup > 0)] -= 1
        else:
            pe = h5.get_node("/Protein/Entries")
            lookup = numpy.zeros(len(pe), dtype="i4")
            for row in pe.where("AltSpliceVariant > 0"):
                lookup[row.nrow] = row["AltSpliceVariant"] - 1
        self.lookup = lookup
        self.n = len(lookup)

    def iter_genes(self):
        taken = numpy.zeros(self.n, dtype=bool)
        for i in range(self.n):
            if taken[i]:
                continue
            if self.lookup[i] == 0:
                taken[i] = True
                yield GeneEntries(numpy.array([i + 1], dtype="i4"), i + 1)
            else:
                main = self.lookup[i]
                l = self.lookup[i : i + 1000]
                res = numpy.where(l == main)[0] + i
                taken[res] = True
                yield GeneEntries(res + 1, main + 1)


class KeywordIndexer:
    def __init__(self, stopwords: Union[os.PathLike, Set[str]]):
        self.kw = collections.defaultdict(set)
        if isinstance(stopwords, set):
            self.stopwords = set(stopwords)
        else:
            self.stopwords = set()
            with open(stopwords, "rt") as fh:
                for line in fh:
                    self.stopwords.union(line.split())

    def process(self, desc, enr):
        pass


class XRefIndexHandler(BaseProfileBuilderProcess):
    def __init__(self, outfile, **kwargs):
        super().__init__(**kwargs)
        if outfile is None:
            outfile = Path(os.getenv("TMPDIR", "/tmp")) / "tmp_index.h5"
        self.outfile = outfile
        self.tmp_h5 = outfile + ".tmp"
        self.kwi = None

    def setup(self):
        self.xref_h5 = tables.open_file(self.tmp_h5, "w", filters=tables.Filters(6, complib="blosc"))
        xref_dtype = numpy.dtype([("XRefId", "S50"), ("EntryNr", "i4"), ("XRefRow", "i4")])
        grp = self.xref_h5.create_group("/", "XRefIndex", title="auxiliary lookup tables with deduplicated xrefs")
        self.xref_idx = self.xref_h5.create_table(
            grp,
            "XRefs",
            description=xref_dtype,
            chunkshape=(16384,),
        )
        self.genenames = collections.defaultdict(list)
        self.spids = collections.defaultdict(list)
        self._buffer = []
        self.kwi = KeywordIndexer()

    def _sort_and_store_xrefs(self):
        data = self.xref_idx.read()
        data.sort(order=["XRefId", "EntryNr"])
        self.xref_h5.close()
        self.xref_h5 = tables.open_file(self.outfile, "w", filters=tables.Filters(6, complib="blosc"))
        xref_idx = self.xref_h5.create_table(
            "/XRefIndex", "XRefs", obj=data, expectedrows=len(data), createparents=True
        )
        xref_idx.colinstances["XRefId"].create_csindex()
        self.xref_idx = xref_idx

    def _store_names(self, dic, node, skip_short=0):
        keys = sorted(dic.keys())
        max_key_len = max(len(z) for z in keys)
        dtyp_key = numpy.dtype([("XRefId", "S{}".format(max_key_len)), ("Offset", "i4"), ("Length", "i4")])
        dtyp_lookup = numpy.dtype([("EntryNr", "i4"), ("XRefRow", "i4")])
        tab_key = self.xref_h5.create_table("/XRefIndex", node, description=dtyp_key, expectedrows=len(keys))
        tab_lookup = self.xref_h5.create_table(
            "/XRefIndex",
            node + "_lookup",
            description=dtyp_lookup,
            expectedrows=sum(len(z) for z in dic.values()),
        )
        off, skipped = 0, 0
        for k in keys:
            if len(dic[k]) < skip_short:
                logger.debug("skipping %s from index", k)
                skipped += 1
                continue
            tab_lookup.append(sorted(dic[k]))
            tab_key.append([(k, off, len(dic[k]))])
            off += len(dic[k])
        tab_lookup.flush()
        tab_key.flush()
        if len(tab_lookup) != off:
            raise Exception("Lookup table length does not match key table offsets")
        tab_key.colinstances["XRefId"].create_csindex()
        logger.info("stored {} index with {} elements. skipped {} keys".format(node, len(tab_key), skipped))

    def _add_to_buffer(self, e):
        self._buffer.append(e)
        if len(self._buffer) >= 2 * self.xref_idx.chunkshape[0]:
            self._flush()

    def _flush(self):
        self.xref_idx.append(self._buffer)
        self._buffer = []

    def add_xref(self, xref, enr, xref_row):
        xref = xref.lower()
        self._add_to_buffer((xref, enr, xref_row))

    def add_swissprot(self, id, enr, xref_row):
        self.spids[id.lower()].append((enr, xref_row))
        self.add_xref(id, enr, xref_row)

    def add_gene_name(self, id, enr, xref_row):
        self.genenames[id.lower()].append((enr, xref_row))
        self.add_xref(id, enr, xref_row)

    def add_keyword(self, enr: int, desc: str):
        self.kwi.process(desc, enr)

    def handle_input(self, item: Tuple[pandas.DataFrame, List]):
        df, desc = item
        for row in df.to_records(index=False):
            if row["XRefSource"] == 0:
                self.add_swissprot(row["XRefId"], row["EntryNr"], row["xref_row"])
            elif row["XRefSource"] in (110, 115):
                self.add_gene_name(row["XRefId"], row["EntryNr"], row["xref_row"])
            else:
                self.add_xref(row["XRefId"], row["EntryNr"], row["xref_row"])

    def finalize(self):
        self._flush()
        self._sort_and_store_xrefs()
        self._store_names(self.spids, "SwissProt")
        self._store_names(self.genenames, "GeneNames", 3)
        self.xref_h5.close()
        os.remove(self.tmp_h5)


RE_VERS = re.compile(rb"(?P<base>[a-z0-9_.-]+)\.\d{1,2}$")


def rem_vers(x):
    m = RE_VERS.match(x)
    if m is not None:
        return m.group("base")
    return x


class GeneGenerator(SourceProcess):
    def __init__(self, h5_path, **kwargs):
        super().__init__(**kwargs)
        self.h5_path = h5_path
        self.h5 = None
        self.splice_helper = None

    def setup(self):
        self.h5 = tables.open_file(self.h5_path)
        self.splice_helper = SpliceVariantHelper(self.h5)

    def generate_data(self):
        for gene in self.splice_helper.iter_genes():
            yield gene

    def finalize(self):
        self.h5.close()


class XRefReducer(BaseProfileBuilderProcess):
    def __init__(self, h5_path, **kwargs):
        super().__init__(**kwargs)
        self.h5_path = h5_path
        self.h5 = None
        self.xref_tab = None
        self.xref_eof = None

    def setup(self):
        self.h5 = tables.open_file(self.h5_path)
        self.xref_tab = self.h5.get_node("/XRef")
        try:
            self.xref_eof = self.h5.get_node("/XRef_EntryNr_offset").read()
        except tables.NoSuchNodeError:
            pass

    def finalize(self):
        self.h5.close()

    def _load_xrefs_with_entry_offset(self, gene):
        xrefs = pandas.DataFrame(
            numpy.hstack(
                list(
                    map(
                        lambda s: self.xref_tab[self.xref_eof[s.start] : self.xref_eof[s.stop]],
                        gene.entrynr_slices(),
                    )
                )
            )
        )
        xrefs["xref_row"] = pandas.Series(
            itertools.chain.from_iterable(
                map(
                    lambda s: range(self.xref_eof[s.start], self.xref_eof[s.stop]),
                    gene.entrynr_slices(),
                )
            )
        )
        return xrefs

    def _load_xrefs_with_where_cond(self, gene):
        query = " | ".join(
            ["((EntryNr >= {}) & (EntryNr < {}))".format(s.start, s.stop) for s in gene.entrynr_slices()]
        )
        dtype = [("xref_row", "i8")] + self.xref_tab.dtype.descr
        buf = io.BytesIO()
        for row in self.xref_tab.where(query):
            buf.write(row.nrow.tobytes())
            buf.write(row.fetch_all_fields().tobytes())
        data = numpy.frombuffer(buf.getbuffer(), dtype=dtype)
        return pandas.DataFrame(data)

    def _load_descriptions(self, gene):
        query = " | ".join([f"((EntryNr >= {s.start}) & (EntryNr < {s.stop}))" for s in gene.entrynr_slices()])
        descriptions = []
        desc_buf = self.h5.get_node("/Protein/DescriptionBuffer")
        for row in self.h5.get_node("/Protein/Entries").where(query):
            desc = (
                desc_buf[row["DescriptionOffset"] : row["DescriptionOffset"] + row["DescriptionLength"]]
                .tobytes()
                .decode()
            )
            descriptions.append((row["EntryNr"], desc))
        return descriptions

    def handle_input(self, gene):
        if self.xref_eof is None:
            xrefs = self._load_xrefs_with_where_cond(gene)
        else:
            xrefs = self._load_xrefs_with_entry_offset(gene)
        xrefs["is_main"] = xrefs["EntryNr"] == gene.main

        # transformations
        xrefs["XRefId"] = xrefs["XRefId"].apply(bytes.lower).apply(rem_vers)

        # sort such that first element per xrefid is the one we want to keep in the index
        sorted_xrefs = xrefs.sort_values(["Verification", "is_main", "XRefSource"], ascending=[True, False, True])
        res = sorted_xrefs.groupby("XRefId").first().reset_index()
        desc = self._load_descriptions(gene)
        return res, desc


def reduce_xrefs(h5_path, outpath=None, nr_procs=None):
    pipeline = Pipeline()
    if nr_procs is None:
        nr_procs = multiprocessing.cpu_count()

    pipeline.add_stage(Stage(GeneGenerator, nr_procs=1, h5_path=h5_path))
    pipeline.add_stage(Stage(XRefReducer, nr_procs=nr_procs, h5_path=h5_path))
    pipeline.add_stage(Stage(XRefIndexHandler, nr_procs=1, outfile=outpath))
    print("setup pipeline, about to start it")
    pipeline.run()
    print("finished computing a reduced set of xrefs")



import collections
import operator
import os
import logging
import math
import re
import ftplib

from tqdm import tqdm
from lxml import etree
from .db import Database

logger = logging.getLogger(__name__)

"""Module to generate external data and crosslinks for NCBI.

"""


class NCBILinkOutXML(object):
    root_node = "not_set"
    provider_id = "9822"

    def __init__(self):
        root = etree.Element(self.root_node)
        for key, value in self.root_children().items():
            root.append(self.text_elemement(key, value))
        self.tree = etree.ElementTree(root)
        self._add_doctype()

    def root_children(self):
        return {}

    def _add_doctype(self):
        self.tree.docinfo.public_id = "-//NLM//DTD LinkOut 1.0//EN"
        self.tree.docinfo.system_url = "https://www.ncbi.nlm.nih.gov/projects/linkout/doc/LinkOut.dtd"

    def text_elemement(self, tag, text):
        el = etree.Element(tag)
        el.text = text
        return el

    def write(self, fh):
        fh.write(etree.tostring(self.tree, pretty_print=True, xml_declaration=True, encoding="utf-8"))


class Provider(NCBILinkOutXML):
    root_node = "Provider"

    def root_children(self):
        elements = collections.OrderedDict(
            [
                ("ProviderId", self.provider_id),
                ("Name", "OMA Browser: Orthologous MAtrix"),
                ("NameAbbr", "OMA"),
                # ("SubjectType", "taxonomy/phylogenetic"),
                ("Url", "http://omabrowser.org"),
                (
                    "Brief",
                    "OMA is a method and database for the inference of orthologs among complete genomes. "
                    "We provide browsable orthology predictions, APIs, flat file downloads among thousands "
                    "of genomes.",
                ),
            ]
        )
        return elements


class Resource(NCBILinkOutXML):
    root_node = "LinkSet"
    link_id = 1

    def _add_objs(self, accs):
        objsel = etree.Element("ObjectSelector")
        objsel.append(self.text_elemement("Database", self.database()))
        objlst = etree.Element("ObjectList")
        objsel.append(objlst)
        for acc in accs:
            objlst.append(self.object_node(acc))
        return objsel

    def object_node(self, acc):
        return self.text_elemement("ObjId", acc)

    def _add_url_section(self, acc):
        el = etree.Element("ObjectUrl")
        el.append(self.text_elemement("Base", self.base_url()))
        nxt = rule = etree.Element("Rule")
        for k, rule_part in enumerate(self.rule_url(acc)):
            if isinstance(rule_part, str):
                if k == 0:
                    nxt.text = rule_part
                else:
                    nxt.tail = rule_part
            elif rule_part.tag == etree.Entity:
                nxt.append(rule_part)
                nxt = rule_part
        el.append(rule)
        el.append(self.text_elemement("SubjectType", "taxonomy/phylogenetic"))
        return el

    def add_link(self, accs):
        lnk = etree.Element("Link")
        lnk.append(self.text_elemement("LinkId", str(self.link_id)))
        lnk.append(self.text_elemement("ProviderId", self.provider_id))
        lnk.append(self._add_objs(accs))
        lnk.append(self._add_url_section(accs))
        self.tree.getroot().append(lnk)
        self._bump_link_id()

    @classmethod
    def _bump_link_id(cls):
        cls.link_id += 1

    def database(self):
        return "not set"

    def base_url(self):
        return "https://omabrowser.org/oma/hogs/"

    def rule_url(self, acc):
        return ("",)


class GenesResource(Resource):
    DISKSIZE_HEADER = 200
    DISKSIZE_PER_LINK = 435
    base_name = "resource_genes"

    def base_url(self):
        return "https://omabrowser.org/oma/info/"

    def rule_url(self, acc):
        return (next(iter(acc.values())) + "/",)

    def database(self):
        return "Gene"


class ProteinResource(Resource):
    DISKSIZE_HEADER = 500
    DISKSIZE_PER_LINK = 45
    base_name = "resource_protein"

    def base_url(self):
        return "https://omabrowser.org/oma/hogs/"

    def object_node(self, acc):
        return self.text_elemement("Query", "{}[accn]".format(acc))

    def rule_url(self, acc):
        return etree.Entity("lo.pacc"), "/"

    def database(self):
        return "Protein"


class TaxonomyResource(Resource):
    DISKSIZE_HEADER = 200
    DISKSIZE_PER_LINK = 435
    base_name = "resource_taxonomy"

    def database(self):
        return "taxonomy"

    def base_url(self):
        return "https://omabrowser.org/oma/genome/"

    def rule_url(self, acc):
        return (next(iter(acc.values())) + "/info/",)


class AncestralTaxonomyResource(TaxonomyResource):
    def base_url(self):
        return "https://omabrowser.org/oma/ancestralgenome/"

    @classmethod
    def set_link_offset(cls, off):
        cls.link_id = off


class LinkoutBuffer(object):
    def __init__(self, resource, outdir="/tmp", bulk_add=True, max_file_size=20 * 2**20):
        self.max_records = math.floor((max_file_size - resource.DISKSIZE_HEADER) / resource.DISKSIZE_PER_LINK)
        self.cur_nr = 0
        self.buf = []
        self.bulk_add = bulk_add
        self.resource_type = resource
        self.outdir = outdir
        logger.info(
            "Setup Linkout buffer for {} with max {} records ({}bytes) per file, bulk_add={}".format(
                resource.__name__, self.max_records, max_file_size, bulk_add
            )
        )

    def add(self, obj):
        self.buf.append(obj)
        if len(self.buf) >= self.max_records:
            self.flush()

    def flush(self):
        res = self.resource_type()
        if self.bulk_add:
            res.add_link(self.buf)
        else:
            for obj in self.buf:
                res.add_link(obj)
        fn = os.path.join(self.outdir, "{}_{:02d}.xml".format(res.base_name, self.cur_nr))
        with open(fn, "wb") as fh:
            res.write(fh)
        self.cur_nr += 1
        self.buf = []


class GenesPriorizationHandler(object):
    """Adapter to LinkoutBuffer to select only a limited number of crossrefs.
    NCBI linkout caps at 10%"""

    def __init__(self, max_linkouts=None, db=None, **kwargs):
        self.max_links = int(max_linkouts) if max_linkouts else 20357436 // 10  # obtained in Jan2018
        logger.info("Limiting Genes to %d links max", self.max_links)
        self.genes_buffer = LinkoutBuffer(GenesResource, **kwargs)
        self.genes = []
        self.db = db

    def add(self, key, value):
        self.genes.append((key, value))

    def _genome_size_map(self):
        gs = self.db.get_hdf5_handle().get_node("/Genome").read()
        return {row["UniProtSpeciesCode"].decode(): row["TotEntries"] for row in gs}

    def flush(self):
        priority_prefixes = [
            "HUMAN",
            "MOUSE",
            "RATNO",
            "PIGXX",
            "DRO",
            "SCH",
            "YEAST",
            "ARA",
            "WHEAT",
            "PLAF",
            "ECO",
            "BAC",
            "PANTR",
            "ORY",
            "GOSHI",
            "BRA",
            "DANRE",
            "CAE",
            "MYC",
            "STR",
            "MAIZE",
            "GORGO",
            "PANTR",
            "PONAB",
            "MACMU",
            "YARLI",
            "PEDHC",
            "TRICA",
            "XENTR",
            "YERPE",
            "POPTR",
        ]
        pat = re.compile(r"^({})".format("|".join(priority_prefixes)))
        if len(self.genes) > self.max_links:
            # final sort order will be 'priority genome', genome size and proteins within genome
            self.genes.sort(key=operator.itemgetter(1))
            if self.db is not None:
                genome_size = self._genome_size_map()
                self.genes.sort(key=lambda x: genome_size[x[1][0:5]], reverse=True)
            self.genes.sort(key=lambda x: pat.match(x[1]) is None)
        for link_acc, link_target in self.genes[0 : self.max_links]:
            self.genes_buffer.add({link_acc: link_target})
        self.genes_buffer.flush()

        c = collections.defaultdict(int)
        for acc, target in self.genes[self.max_links :]:
            c[target[0:5]] += 1
        logger.info("Skipping genes link in the following species: %s", c)


def prepare_linkout_files(outdir="/tmp", infile="../pyomabrowser/OmaServer.h5"):
    prov = Provider()
    with open(os.path.join(outdir, "providerinfo.xml"), "wb") as fh:
        prov.write(fh)

    db = Database(infile)
    xrefs = db.get_hdf5_handle().get_node("/XRef")
    xref_source_enum = xrefs.get_enum("XRefSource")

    protein_buffer = LinkoutBuffer(ProteinResource, outdir=outdir, bulk_add=True)
    genes_buffer = GenesPriorizationHandler(db=db, outdir=outdir, bulk_add=False)
    for xref in tqdm(xrefs):
        if xref["XRefSource"] == xref_source_enum["RefSeq"]:
            protein_buffer.add(xref["XRefId"].decode())
        elif xref["XRefSource"] == xref_source_enum["EntrezGene"]:
            genes_buffer.add(
                xref["XRefId"].decode(),
                db.id_mapper["OMA"].map_entry_nr(xref["EntryNr"]),
            )
    protein_buffer.flush()
    genes_buffer.flush()

    with open(os.path.join(outdir, "resource_taxonomy_00.xml"), "wb") as fh:
        taxs = TaxonomyResource()
        for row in db.id_mapper["OMA"].genome_table:
            taxs.add_link({str(row["NCBITaxonId"]): row["UniProtSpeciesCode"].decode()})
        taxs.write(fh)
    # add Ancestral genomes links
    with open(os.path.join(outdir, "resource_taxonomy_01.xml"), "wb") as fh:
        anctaxs = AncestralTaxonomyResource()
        anctaxs.set_link_offset(taxs.link_id)
        for row in db.tax.tax_table:
            if row["Name"] in db.tax.all_hog_levels:
                anctaxs.add_link({str(row["NCBITaxonId"]): str(row["NCBITaxonId"])})
        anctaxs.write(fh)


def copy_to_ncbi(dir, password, host="ftp-private.ncbi.nlm.nih.gov", user="omabrow"):
    with ftplib.FTP(host, user, password) as session:
        session.cwd("/holdings")

        for fname in os.listdir(dir):
            if fname.endswith(".xml"):
                with open(os.path.join(dir, fname), "rb") as fh:
                    cmd = "STOR {}".format(fname)
                    session.storbinary(cmd, fp=fh)
                logger.info("finished transfering " + fname)



import hashlib
import math
import os
import logging

logger = logging.getLogger(__name__)


class JobArray(object):
    def __init__(self, nr_procs, this_proc_nr):
        self.nr_procs = int(nr_procs)
        self.this_proc_nr = int(this_proc_nr)
        if self.nr_procs < 1 or self.this_proc_nr < 1 or self.this_proc_nr > self.nr_procs:
            raise ValueError(
                "cannot determine HPC parallel job ids: nr_procs={}, this_proc_nr={}".format(nr_procs, this_proc_nr)
            )

    def is_my_job(self, chunk):
        chunk = str(chunk).encode("utf-8") if not isinstance(chunk, bytes) else chunk
        h = hashlib.md5(chunk)
        as_int = int(h.hexdigest(), 16)
        res = (as_int % self.nr_procs) == (self.this_proc_nr - 1)
        logger.debug("chunk %s to be processed: %s", chunk, res)
        return res

    def __repr__(self):
        return "{}({},{})".format(self.__class__.__name__, self.nr_procs, self.this_proc_nr)

    def __str__(self):
        return "Jobarray process {} of {}".format(self.this_proc_nr, self.nr_procs)

    def modify_filename(self, basename):
        if self.nr_procs > 1:
            base, ext = os.path.splitext(basename)
            nr_digits = math.floor(math.log10(self.nr_procs)) + 1
            this_proc = f"{self.this_proc_nr}".zfill(nr_digits)
            return f"{base}_{this_proc}-{self.nr_procs}{ext}"
        return basename


def detect_hpc_jobarray(nr_procs, this_proc_nr=None):
    if this_proc_nr is not None:
        return JobArray(nr_procs, int(this_proc_nr))
    if os.getenv("LSB_JOBID"):
        return JobArray(nr_procs, os.getenv("LSB_JOBINDEX", "1"))
    elif os.getenv("SGE_TASK_ID"):
        return JobArray(nr_procs, os.getenv("SGE_TASK_ID"))
    elif os.getenv("SLURM_ARRAY_JOB_ID"):
        return JobArray(nr_procs, os.getenv("SLURM_ARRAY_TASK_ID"))
    elif os.getenv("THIS_PROC_NR"):
        return JobArray(nr_procs, os.getenv("THIS_PROC_NR"))
    elif nr_procs is None or nr_procs == 1:
        return JobArray(1, 1)
    else:
        raise ValueError("Cannot identify HPC jobarray setup")



import abc
import collections
import itertools
import logging
import re
from typing import Union, List
from dataclasses import dataclass

from .exceptions import (
    InvalidId,
    UnknownSpecies,
    InvalidTaxonId,
    AmbiguousID,
    TooUnspecificQuery,
)
from . import db, models

logger = logging.getLogger(__name__)


class BaseSearch(metaclass=abc.ABCMeta):
    PRIO = 100

    def __init__(self, pyomadb: db.Database, term: Union[int, str]):
        self.db = pyomadb
        self.term = term

    def set_max_entries(self, max_nr: int):
        pass

    def search_entries(self):
        pass

    def search_groups(self):
        pass

    def search_species(self):
        pass

    def search_ancestral_genomes(self):
        pass

    def count_entries(self):
        nr_entries = self.search_entries()
        return 0 if nr_entries is None else len(nr_entries)

    def count_groups(self):
        nr_groups = self.search_groups()
        return 0 if nr_groups is None else len(nr_groups)

    def count_species(self):
        nr_species = self.search_species()
        return 0 if nr_species is None else len(nr_species)

    def count_ancestral_genomes(self):
        nr_ancestral_genomes = self.search_ancestral_genomes()
        return 0 if nr_ancestral_genomes is None else len(nr_ancestral_genomes)


class BaseWithEntryFilter(BaseSearch):
    def __init__(self, pyomadb: db.Database, term: Union[int, str]):
        super().__init__(pyomadb, term)
        self._max_entries = None
        self._entry_nr_filter = None

    def set_max_entries(self, max_nr: int):
        self._max_entries = max_nr

    def set_entry_nr_filter(self, entry_filter: Union[set, tuple]):
        self._entry_nr_filter = entry_filter


class OmaGroupSearch(BaseSearch):
    PRIO = 50

    @models.LazyProperty
    def _matched_groups(self):
        try:
            grps = [self.db.resolve_oma_group(self.term)]
        except AmbiguousID as e:
            grps = e.candidates
        except InvalidId:
            grps = []
        return grps

    def search_groups(self):
        return [models.OmaGroup(self.db, gr) for gr in self._matched_groups]

    def search_entries(self):
        res = []
        for grp in self._matched_groups:
            for en in self.db.oma_group_members(grp):
                res.append(models.ProteinEntry(self.db, en))
        return res


class HogIDSearch(BaseSearch):
    PRIO = 70

    def __init__(
        self,
        pyomadb: db.Database,
        term: Union[int, str],
        level: Union[None, int] = None,
    ):
        super().__init__(pyomadb, term)
        self.outdated_query_hog = False
        self._level = None
        self._get_protein_entries = False
        self._matched_hogs = None
        if level is not None:
            try:
                self.set_taxon_filter(models.AncestralGenome(self.db, level))
            except UnknownSpecies:
                pass

    def set_taxon_filter(self, taxon):
        if not isinstance(taxon, models.AncestralGenome):
            return
        if self._level != taxon.scientific_name:
            self._level = taxon.scientific_name
            self._get_protein_entries = len(taxon.extant_genomes) < 3
            self._matched_hogs = None
            self.outdated_query_hog = False

    def _map_forward_outdated_hogid(self, hogid):
        try:
            with db.HogIdForwardMapper(self.db) as fwd:
                return fwd.map_hogid(hogid)
        except IOError as e:
            return {}

    def get_matched_hogs(self):
        if self._matched_hogs is not None:
            return self._matched_hogs

        match = re.match(
            r"(?P<id>(?P<prefix>HOG:)?(?P<rel>[A-Z]*)(?P<fam>\d+)(?P<subid>[a-z0-9.]*))(?:_(?P<taxid>\d+))?",
            self.term,
        )
        hogs = []
        if match is not None:
            level = None
            logger.debug("analyse token %s: %s", self.term, match.groupdict())
            if match.group("taxid") is not None:
                try:
                    n = self.db.tax.get_taxnode_from_name_or_taxid(match.group("taxid"))
                    level = n[0]["Name"].decode()
                except (InvalidTaxonId, KeyError):
                    pass
            new_hogs = {}
            if match.group("rel") != self.db.release_char:
                old_hogid = ("" if match.group("prefix") else "HOG:") + match.group("id")
                new_hogs = self._map_forward_outdated_hogid(old_hogid)
                self.outdated_query_hog = True
                ids = new_hogs.keys()
            else:
                ids = (
                    ("" if match.group("prefix") else "HOG:") + match.group("id"),
                    "HOG:{}{:07d}".format(self.db.release_char, int(match.group("fam"))),
                )
            if self._level is not None:
                # we have a taxon filter, search for subhogs/superhogs of ids
                ids_to_look = ids if self.outdated_query_hog else (ids[0],)
                for hog_id in ids_to_look:
                    for subhog in self.db.iter_hogs_at_level(hog_id, level=self._level):
                        hogs.append(models.HOG(self.db, subhog))
            levels = [level]
            if level is not None:
                levels.append(None)
            for hog_id, lev in itertools.product(ids, levels):
                try:
                    hog = models.HOG(self.db, self.db.get_hog(hog_id, lev))
                    hogs.append(hog)
                    if self.outdated_query_hog:
                        hog.query_jaccard_similarity = new_hogs[hog_id]
                    else:
                        break  # not an outdated hog. break at the best candidate
                except ValueError:
                    pass
        self._matched_hogs = hogs
        return hogs

    def search_entries(self):
        if not self._get_protein_entries:
            return None
        return list(
            itertools.chain.from_iterable(hog.members for hog in self.get_matched_hogs() if hog.level == self._level)
        )

    def search_groups(self):
        return self.get_matched_hogs()


class GOSearch(BaseWithEntryFilter):
    PRIO = 30

    @models.LazyProperty
    def _matched_entries(self):
        try:
            if self._max_entries is None and self.estimated_occurrences > 1000:
                self.set_max_entries(1000)
            res = list(
                self.db.entrynrs_with_go_annotation(
                    self.term,
                    limit=self._max_entries,
                    entrynr_filter=self._entry_nr_filter,
                )
            )
        except InvalidId:
            res = []
        return res

    def search_entries(self):
        return [models.ProteinEntry(self.db, en) for en in self._matched_entries]

    @models.LazyProperty
    def estimated_occurrences(self):
        return self.db.count_go_annotations(self.term)

    def count_entries(self):
        return self.estimated_occurrences


class ECSearch(BaseWithEntryFilter):
    PRIO = 25

    @models.LazyProperty
    def _matched_entries(self):
        if self._max_entries is None and self.estimated_occurrences > 1000:
            self.set_max_entries(1000)
        return list(
            int(z)
            for z in self.db.entrynrs_with_ec_annotation(
                self.term, limit=self._max_entries, entrynr_filter=self._entry_nr_filter
            )
        )

    def search_entries(self):
        return [models.ProteinEntry(self.db, en) for en in self._matched_entries]

    @models.LazyProperty
    def estimated_occurrences(self):
        return self.db.count_ec_annotations(self.term)

    def count_entries(self):
        return self.estimated_occurrences


class DomainSearch(BaseWithEntryFilter):
    PRIO = 40

    @models.LazyProperty
    def _matched_entries(self):
        if self._max_entries is None and self.estimated_occurrences > 1000:
            self.set_max_entries(1000)
        return list(
            int(z)
            for z in self.db.entrynrs_with_domain_id(
                self.term, limit=self._max_entries, entrynr_filter=self._entry_nr_filter
            )
        )

    def search_entries(self):
        return [models.ProteinEntry(self.db, en) for en in self._matched_entries]

    @models.LazyProperty
    def estimated_occurrences(self):
        return self.db.count_domain_id_annotations(self.term)

    def count_entries(self):
        return self.estimated_occurrences


class TaxSearch(BaseSearch):
    PRIO = 10

    @models.LazyProperty
    def _matched_taxons(self):
        try:
            tax_nodes = self.db.tax.get_taxnode_from_name_or_taxid(self.term)
            tax = [(int(z), 1.0) for z in tax_nodes["NCBITaxonId"]]
        except KeyError:
            try:
                genome = self.db.id_mapper["OMA"].identify_genome(self.term)
                tax = [(int(genome["NCBITaxonId"]), 1.0)]
            except UnknownSpecies:
                # fuzzy match of all extant genomes in OMA
                genomes, genome_approx_scores = self.db.id_mapper["OMA"].approx_search_genomes(self.term, scores=True)
                logger.debug("%s matches approximately to %s", self.term, genomes)
                # fuzzy match of all taxonomic names in OMA.
                approx_matches = self.db.tax.approx_search(self.term)
                logger.debug("'%s' matches approximately to %s", self.term, approx_matches)
                tax_nodes = self.db.tax.get_taxnode_from_name_or_taxid([z[1] for z in approx_matches])
                tax = [(int(z["NCBITaxonId"]), approx_match[0]) for z, approx_match in zip(tax_nodes, approx_matches)]
                tax.extend((g.ncbi_taxon_id, score) for g, score in zip(genomes, genome_approx_scores))
        return tax

    def search_entries(self):
        enrs = set([])
        for g in self.search_species():
            enrs.update(range(g.entry_nr_offset + 1, g.entry_nr_offset + len(g) + 1))
        return enrs

    def search_ancestral_genomes(self):
        ancestral_genomes = []
        for tax, score in self._matched_taxons:
            ag = models.AncestralGenome(self.db, tax)
            if len(ag.extant_genomes) <= 1:
                continue
            if ag.sciname not in self.db.tax.all_hog_levels:
                orig_sciname = ag.sciname
                while True:
                    childs = self.db.tax._direct_children_taxa(int(ag.ncbi_taxon_id))
                    if len(childs) == 0 or len(childs) > 1:
                        break
                    ag = models.AncestralGenome(self.db, int(childs[0]["NCBITaxonId"]))
                ag.common_name = f"Representative for '{orig_sciname}'"
            ag.match_score = score
            ancestral_genomes.append(ag)
        return ancestral_genomes

    def search_species(self):
        extant_species = {}
        for tax, score in self._matched_taxons:
            extant_species_of_tax = models.AncestralGenome(self.db, tax).extant_genomes
            if len(extant_species_of_tax) > 1:
                score /= 2
            for sp in extant_species_of_tax:
                if sp.ncbi_taxon_id not in extant_species:
                    sp.match_score = score
                    extant_species[sp.ncbi_taxon_id] = sp
                else:
                    genome = extant_species[sp.ncbi_taxon_id]
                    genome.match_score = max(score, genome.match_score)
        return list(extant_species.values())


class SequenceSearch(BaseSearch):
    PRIO = 90

    def __init__(
        self,
        pyomadb: db.Database,
        term: str,
        strategy: Union[None, str] = None,
    ):
        super().__init__(pyomadb, term)
        self.strategy = strategy.lower() if strategy else "mixed"
        self.seq = self.db.seq_search._sanitise_seq(self.term)
        self.entry_filter = None
        self._matched_seqs = None
        self._nr_kmer_hits = None
        self._max_entries = None

    def set_entry_nr_filter(self, filter: Union[tuple, set]):
        if isinstance(filter, tuple) and len(filter) != 2:
            raise ValueError("filter parameter must be range tuple or a set")
        self.entry_filter = filter
        self._matched_seqs = None
        self._nr_kmer_hits = None

    def set_max_entries(self, max_nr: int):
        self._max_entries = max_nr

    def get_matched_seqs(self):
        if self._matched_seqs is not None:
            return self._matched_seqs

        if len(self.seq) < 5:
            logger.debug("too short sequence motif to search: %s", self.seq)
            raise ValueError("too short sequence motif")
        if self.strategy not in ("exact", "approx", "mixed"):
            raise ValueError("Invalid search strategy parameter")

        res = {}
        if self.strategy in ("exact", "mixed"):
            exact_matches = self.db.seq_search.exact_search(
                self.seq,
                only_full_length=False,
                is_sanitised=True,
                entrynr_range=self.entry_filter,
            )
            for en in exact_matches:
                pe = models.ProteinEntry(self.db, en)
                pe.mode = "exact"
                pe.score = 5 * len(self.seq)
                pe.alignment = self.seq
                res[en] = pe

            self._nr_kmer_hits = len(exact_matches)

        if self.strategy == "approx" or self.strategy == "mixed" and len(res) == 0:
            approx, kmer_hits = self.db.seq_search.approx_search(
                self.seq,
                n=self._max_entries,
                is_sanitised=True,
                entrynr_range=self.entry_filter,
                return_kmer_hits=True,
            )
            for en, align_res in approx:
                pe = models.ProteinEntry(self.db, en)
                pe.mode = "approx"
                pe.score = align_res["score"]
                pe.alignment = align_res["alignment"]
                res[en] = pe

            self._nr_kmer_hits = len(kmer_hits)

        self._matched_seqs = res
        return res

    def search_entries(self):
        return list(self.get_matched_seqs().values())

    def search_groups(self):
        grps = collections.Counter(p.oma_group for p in self.get_matched_seqs().values() if p.oma_group != 0)
        return [models.OmaGroup(self.db, grp) for grp, cnt in grps.most_common(10)]

    def count_entries(self):
        if self._matched_seqs is None:
            self.get_matched_seqs()
        if self._nr_kmer_hits is None:
            return len(self._matched_seqs)
        else:
            return self._nr_kmer_hits


class XRefSearch(BaseSearch):
    PRIO = 85

    def __init__(
        self,
        pyomadb: db.Database,
        term: str,
        max_matches: Union[None, int] = None,
    ):
        super().__init__(pyomadb, term)
        self.max_matches = max_matches
        self._matched_entries = None
        self.entry_filter = None

    def set_max_entries(self, max_nr: int):
        self.max_matches = max_nr

    @models.LazyProperty
    def estimated_occurrences(self):
        return self.db.id_mapper["XRef"].search_helper.count(self.term)

    def count_entries(self):
        return self.estimated_occurrences

    def set_entry_nr_filter(self, enr_filter: Union[tuple, set]):
        if isinstance(enr_filter, tuple) and len(enr_filter) > 2:
            raise ValueError("enr_filter parameter must be range tuple or a set")
        self.entry_filter = enr_filter
        self._matched_entries = None

    def get_matched_proteins(self):
        if self._matched_entries is None:
            filt = None
            rng = None
            if self.entry_filter is not None:
                if isinstance(self.entry_filter, tuple):
                    rng = self.entry_filter
                else:
                    if len(self.entry_filter) == 0:
                        logger.info("short-circuit XRefSearch as entry_filter is empty: %s", self.entry_filter)
                        self._matched_entries = {}
                        return self._matched_entries
                    rng = (min(self.entry_filter), max(self.entry_filter))
                    filt = lambda enr: enr in self.entry_filter

            try:
                self._matched_entries = self.db.id_resolver.search_protein(
                    self.term, limit=self.max_matches, entrynr_range=rng
                )
            except TooUnspecificQuery:
                logger.exception("XRefSearch with term %s and entry_range %s", self.term, rng)
                self._matched_entries = {}
            if filt is not None:
                self._matched_entries = {enr: v for enr, v in self._matched_entries.items() if filt(enr)}
        return self._matched_entries

    def search_entries(self):
        res = []
        for enr, xrefdata in self.get_matched_proteins().items():
            p = models.ProteinEntry(self.db, enr)
            p.xref_data = xrefdata
            res.append(p)
        return res


@dataclass
class SearchResult:
    entries: dict = None
    groups: dict = None
    species: dict = None
    ancestral_genomes: dict = None
    entries_set: set = None
    groups_set: set = None
    species_set: set = None
    ancestral_genomes_set: set = None

    def __and__(self, other: BaseSearch):
        assert isinstance(other, BaseSearch)
        if self.entries is not None and hasattr(other, "set_entry_nr_filter"):
            other.set_entry_nr_filter(self.entries_set)
        if (
            self.ancestral_genomes_set is not None
            and len(self.ancestral_genomes_set) == 1
            and hasattr(other, "set_taxon_filter")
        ):
            ag = next(iter(self.ancestral_genomes_set))
            other.set_taxon_filter(self.ancestral_genomes[ag])

        res = SearchResult()
        for aspect, key in zip(
            ("entries", "groups", "species", "ancestral_genomes"),
            ("entry_nr", "group_nbr", "ncbi_taxon_id", "ncbi_taxon_id"),
        ):
            term_aspect_result = getattr(other, "search_" + aspect)()
            if term_aspect_result is not None:
                if aspect == "groups" and isinstance(other, HogIDSearch):
                    key = "hog_id"
                if isinstance(term_aspect_result, set):
                    keyset = term_aspect_result
                    keyvals = collections.defaultdict(dict)
                else:
                    keyset = set(getattr(z, key) for z in term_aspect_result)
                    keyvals = {getattr(z, key): z for z in term_aspect_result}

                if getattr(self, aspect) is None:
                    setattr(res, aspect, keyvals)
                    setattr(res, aspect + "_set", keyset)
                else:
                    common_keys = getattr(self, aspect + "_set") & keyset
                    setattr(res, aspect + "_set", common_keys)
                    d = getattr(self, aspect)
                    if len(d) > 0:
                        common_dict = {k: d[k] for k in common_keys}
                        common_dict.update({k: keyvals[k] for k in common_keys})
                    else:
                        common_dict = keyvals
                    setattr(res, aspect, common_dict)
            else:
                setattr(res, aspect + "_set", getattr(self, aspect + "_set"))
                setattr(res, aspect, getattr(self, aspect))
        return res


def search(tokens: List[BaseSearch], entry_limit=None):
    sorted_tokens = sorted(tokens, key=lambda t: t.PRIO)
    if entry_limit is not None:
        sorted_tokens[-1].set_max_entries(entry_limit)
    res = SearchResult()
    for token in sorted_tokens:
        res &= token
    return res



import collections
import functools
import itertools
import json
import logging
import multiprocessing as mp
import os
import signal
import sys
import time
from os.path import commonprefix, split
from queue import Empty

import numpy
import numpy.lib.recfunctions
import tables
from tqdm import tqdm

from .db import Database
from .exceptions import DBConsistencyError
from .models import ProteinEntry
from .tablefmt import ProteinCacheInfo

logger = logging.getLogger(__name__)

Protein = collections.namedtuple("Protein", ("entry_nr", "hog_id", "group"))


def signal_handler(signum, frame):
    logger.info("received signal " + str(signum))
    raise KeyboardInterrupt(signum)


def length_limited_set_formatter(s):
    if len(s) > 10:
        x = s.pop()
        s.add(x)
        return "Set(size={}, ex: {})".format(len(s), x)
    else:
        return str(s)


def are_orthologous(a: Protein, b: Protein):
    if a.entry_nr == b.entry_nr:
        return False
    prefix = commonprefix((a.hog_id, b.hog_id))
    if "." in prefix and prefix[-1].isdigit():
        return False
    return True


class CacheBuilderWorker(mp.Process):
    def __init__(self, db_fpath, in_queue, out_queue, **kwargs):
        super(CacheBuilderWorker, self).__init__(**kwargs)
        self.db_fpath = db_fpath
        self.in_queue = in_queue
        self.out_queue = out_queue
        self.h5 = None
        self.db = None

    def run(self):
        self.db = Database(self.db_fpath)
        self.h5 = self.db.get_hdf5_handle()
        timelimit_get_from_in_queue = functools.partial(self.in_queue.get, timeout=120)

        try:
            try:
                for job in iter(timelimit_get_from_in_queue, None):
                    fun, params = job
                    res = getattr(self, fun)(*params)
                    logger.debug("result for job (%s) ready", job)
                    self.out_queue.put((job, res))
            except Empty:
                logger.warning("No item nor termination signal received in Queue. Giving up")
                logger.exception("Work-queue is empty")
            self.out_queue.put("DONE")
        except KeyboardInterrupt:
            logger.info("received interrupt. Terminating")
        self.db.close()
        logger.info("terminating worker process %s", self.name)

    def load_fam_members(self, fam):
        members = []
        vals = {k: self.db.format_hogid(x).encode("utf-8") for k, x in zip(("fam", "fam_next"), (fam, fam + 1))}
        for row in self.h5.get_node("/Protein/Entries").where(
            "(fam <= OmaHOG) & (OmaHOG < fam_next)",
            condvars=vals,
        ):
            members.append(Protein(row["EntryNr"], row["OmaHOG"].decode(), row["OmaGroup"]))
        return members

    def load_vps(self, entry_nr):
        return self.db.get_vpairs(entry_nr)["EntryNr2"]

    def load_grp_members(self, group):
        return [row["EntryNr"] for row in self.h5.get_node("/Protein/Entries").where("OmaGroup == {:d}".format(group))]

    def analyse_fam(self, fam, rng=None):
        logger.debug("analysing family %s", fam)
        fam_members = self.load_fam_members(fam)
        logger.debug("family {} with {} members; doing range {}".format(fam, len(fam_members), rng))
        grp_members = {grp: set(self.load_grp_members(grp)) for grp in set(z.group for z in fam_members if z.group > 0)}
        nr_memb = len(fam_members)
        fam_iter = iter(fam_members)
        if rng is not None:
            nr_memb = rng[1] - rng[0]
            fam_iter = itertools.islice(fam_members, rng[0], rng[1])

        counts = numpy.zeros(nr_memb, dtype=tables.dtype_from_descr(ProteinCacheInfo))
        for i, p1 in tqdm(
            enumerate(fam_iter),
            disable=len(fam_members) < 500,
            desc="fam {}".format(fam),
            total=nr_memb,
        ):
            vps = set(self.load_vps(p1.entry_nr))
            ind_orth = set(p2.entry_nr for p2 in fam_members if are_orthologous(p1, p2))
            grp = grp_members.get(p1.group, set([])) - set([p1.entry_nr])
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    "entry {}: vps: {} ipw: {} grp: {} any: {}".format(
                        p1.entry_nr,
                        length_limited_set_formatter(vps),
                        length_limited_set_formatter(ind_orth),
                        length_limited_set_formatter(grp),
                        length_limited_set_formatter(vps | ind_orth | grp),
                    )
                )
            counts[i]["EntryNr"] = p1.entry_nr
            counts[i]["NrPairwiseOrthologs"] = len(vps)
            counts[i]["NrHogInducedPWOrthologs"] = len(ind_orth)
            counts[i]["NrHogInducedPWParalogs"] = len(fam_members) - len(ind_orth) - 1
            counts[i]["NrOMAGroupOrthologs"] = len(grp)
            counts[i]["NrAnyOrthologs"] = len(vps | ind_orth | grp)
        return counts

    def analyse_singleton(self, entry_nr, group_nr):
        logger.debug("analysing singleton %s (grp %s)", entry_nr, group_nr)
        vps = set(self.load_vps(entry_nr))
        grp_members = set([])
        if group_nr > 0:
            grp_members = set(self.load_grp_members(group_nr))
        counts = numpy.array(
            [(entry_nr, len(vps), 0, 0, len(grp_members), len(vps | grp_members))],
            dtype=tables.dtype_from_descr(ProteinCacheInfo),
        )
        return counts

    def compute_familydata_json(self, fam):
        famhog_id = self.db.format_hogid(fam)
        logger.debug("analysing family %s", fam)
        fam_members = self.load_fam_members(fam)
        logger.debug("family %s with %d members", fam, len(fam_members))
        if len(fam_members) > 50000:
            # this will likely fail to compute the MDS for so many points
            # let's skip it for now.
            genes_null_similarity = set(p.entry_nr for p in fam_members)
        else:
            try:
                (
                    genes_null_similarity,
                    gene_similarity_vals,
                ) = self.db.get_gene_similarities_hog(famhog_id)
            except Exception as e:
                logger.error("gene_similarity failed for %s: %s", fam, e)
                raise
        final_json_output = []
        for p1 in fam_members:
            to_append = {}
            protein = ProteinEntry(self.db, p1.entry_nr)
            to_append["id"] = p1.entry_nr
            to_append["protid"] = protein.omaid
            to_append["sequence_length"] = protein.sequence_length
            to_append["taxon"] = protein.genome.species_and_strain_as_dict
            to_append["xrefid"] = protein.canonicalid
            to_append["gc_content"] = protein.gc_content
            to_append["nr_exons"] = protein.nr_exons
            if p1.entry_nr in genes_null_similarity:
                to_append["gene_similarity"] = None
            else:
                to_append["gene_similarity"] = gene_similarity_vals[p1.entry_nr]
            final_json_output.append(to_append)

        return json.dumps(final_json_output)


class ResultHandler:
    def __init__(self, cache_file):
        self.cache_file = cache_file
        self.ortholog_count_result = []
        self.family_json_offset = []
        self.in_memory_json_buffer = []
        self.buffer_offset = 0
        self.jobs = []
        self._last_cache_timestamp = time.time()
        self._offset_dtype = [("Fam", "i4"), ("offset", "i8"), ("length", "i4")]

    def load_cache(self):
        def resilient_load_data_node(h5, node):
            try:
                res = [h5.get_node(node).read()]
            except tables.NoSuchNodeError:
                res = []
            return res

        with tables.open_file(self.cache_file) as cache:
            self.ortholog_count_result = resilient_load_data_node(cache, "/ortholog_counts")
            self.family_json_offset = resilient_load_data_node(cache, "/family_json/offset")
            self.buffer_offset = len(cache.root.family_json.buffer)
            self.jobs = cache.root.pending_jobs.read(0)[0]

    def add_jobs(self, jobs):
        self.jobs.extend(jobs)

    def handle_result(self, job, result):
        if job[0] == "compute_familydata_json":
            self.store_familydata_json_result(job, result)
        elif job[0] in ("analyse_fam", "analyse_singleton"):
            self.store_ortholog_count_result(job, result)
        else:
            raise ValueError("Unexpected result type")
        self.jobs.remove(job)
        if time.time() - self._last_cache_timestamp > 300 or sum(len(z) for z in self.in_memory_json_buffer) > 100e6:
            self.write_to_disk()

    def store_familydata_json_result(self, job, result):
        fam = job[1][0]
        encoded_json = result.encode("utf-8")
        json_as_np = numpy.ndarray((len(encoded_json),), buffer=encoded_json, dtype=tables.StringAtom(1))
        self.in_memory_json_buffer.append(json_as_np)
        self.family_json_offset.append(
            numpy.array([(fam, self.buffer_offset, len(encoded_json))], dtype=self._offset_dtype)
        )
        self.buffer_offset += len(encoded_json)

    def store_ortholog_count_result(self, job, result):
        self.ortholog_count_result.append(result)

    def write_to_disk(self):
        logger.info("writing a milestone to disk...")
        transfer_data = False
        if os.path.exists(self.cache_file):
            os.replace(self.cache_file, self.cache_file + ".0")
            transfer_data = True
        with tables.open_file(self.cache_file, "w") as h5:
            buf = h5.create_earray(
                "/family_json",
                "buffer",
                tables.StringAtom(1),
                (0,),
                createparents=True,
                expectedrows=1e9,
            )
            if transfer_data:
                with tables.open_file(self.cache_file + ".0", "r") as prev:
                    buf.append(prev.root.family_json.buffer.read())
            for el in self.in_memory_json_buffer:
                buf.append(el)
            buf.flush()
            if len(self.family_json_offset) > 0:
                off = numpy.lib.recfunctions.stack_arrays(self.family_json_offset, usemask=False)
                h5.create_table("/family_json", "offset", None, obj=off)

            if len(self.ortholog_count_result) > 0:
                cnts = numpy.lib.recfunctions.stack_arrays(self.ortholog_count_result, usemask=False)
                h5.create_table("/", "ortholog_counts", ProteinCacheInfo, obj=cnts)

            a = h5.create_vlarray("/", "pending_jobs", tables.ObjectAtom())
            a.append(self.jobs)
            h5.flush()
            if len(buf) != self.buffer_offset:
                raise DBConsistencyError("buffer has unexpeced length: {}vs{}".format(len(buf), self.buffer_offset))
            self.in_memory_json_buffer = []
            self._last_cache_timestamp = time.time()
            logger.info("finished writing milestone to %s", self.cache_file)


def build_cache(db_fpath, nr_procs=None, from_cache=None):
    request_queue = mp.Queue()
    result_queue = mp.Queue()
    nr_procs = nr_procs if nr_procs else mp.cpu_count()

    db = Database(db_fpath)
    nr_entries = len(db.get_hdf5_handle().get_node("/Protein/Entries"))
    result_handler = ResultHandler(from_cache)
    if from_cache is not None and os.path.isfile(from_cache):
        result_handler.load_cache()
        jobs = result_handler.jobs
        logger.debug(jobs)
    else:
        nr_fams = db.get_nr_toplevel_hogs()
        singletons = [
            (int(r["EntryNr"]), int(r["OmaGroup"]))
            for r in db.get_hdf5_handle().get_node("/Protein/Entries").where('OmaHOG == b""')
        ]

        logger.info("found {} hog and {} singleton jobs to be computed".format(nr_fams, len(singletons)))
        jobs = [("analyse_fam", (fam + 1,)) for fam in range(nr_fams)]
        jobs.extend([("compute_familydata_json", (fam + 1,)) for fam in range(nr_fams)])
        jobs.extend([("analyse_singleton", singleton) for singleton in singletons])
        logger.info("nr of jobs: {} (expected {})".format(len(jobs), 2 * nr_fams + len(singletons)))
        result_handler.add_jobs(jobs)
    db.close()

    workers = []
    for i in range(nr_procs):
        w = CacheBuilderWorker(db_fpath, request_queue, result_queue, daemon=True)
        w.start()
        workers.append(w)

    pending_jobs = set([])
    for job in jobs:
        request_queue.put(job)
        pending_jobs.add(job)
    # Sentinel objects to allow clean shutdown: 1 per worker.
    for i in range(nr_procs):
        request_queue.put(None)

    finished = 0
    try:
        logger.info("start to receive results")
        last_cache_timestamp = time.time()
        for reply in tqdm(iter(result_queue.get, None), total=len(jobs)):
            if reply == "DONE":
                finished += 1
                logger.info("%d workers finished", finished)
                if finished == nr_procs:
                    if len(pending_jobs) > 0:
                        logger.error("still %d pending jobs...: %s", len(pending_jobs), pending_jobs)
                    break
            else:
                job, res = reply
                logger.debug("received result for job %s", job)
                result_handler.handle_result(job, res)
        result_handler.write_to_disk()
        logger.info("exit receiver loop. joining workers...")
        for w in workers:
            w.join()
        logger.debug("all workers joined")
    except KeyboardInterrupt as e:
        logger.info("recived interrupt. writeing out temp results")
        result_handler.write_to_disk()
        sys.exit(99)

    ret = numpy.lib.recfunctions.stack_arrays(result_handler.ortholog_count_result, usemask=False)
    ret.sort(order="EntryNr")
    logger.info("sorted results: %s", ret)
    assert check_all_there(nr_entries, ret)
    return ret


def check_all_there(nr_prots, cache):
    if len(cache) == nr_prots:
        return True
    missings = set(range(1, nr_prots + 1)) - set(cache["EntryNr"])
    logger.error("Missing cache value for %s", missings)
    return False


def compute_and_store_cached_data(db_fpath, nr_procs=None, force=False, tmp_cache=None):
    ortholog_cnts_cache_path = "/Protein/OrthologsCountCache"
    if tmp_cache is None:
        tmp_cache = "/tmp/compute_cache.h5"
    with tables.open_file(db_fpath, "a") as h5:
        try:
            n = h5.get_node(ortholog_cnts_cache_path)
            if force:
                h5.remove_node(ortholog_cnts_cache_path)
            else:
                return
        except tables.NoSuchNodeError:
            pass

    signal.signal(signal.SIGUSR2, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    cache = build_cache(db_fpath, nr_procs=nr_procs, from_cache=tmp_cache)
    update_hdf5_from_cachefile(db_fpath, tmp_cache)


def update_hdf5_from_cachefile(db_fpath, tmp_cache):
    ortholog_cnts_cache_path = "/Protein/OrthologsCountCache"
    path, name = split(ortholog_cnts_cache_path)

    with tables.open_file(db_fpath, "a") as h5, tables.open_file(tmp_cache, "r") as cache_h5:
        cached_cnts = cache_h5.get_node("/ortholog_counts").read()
        cached_cnts.sort(order="EntryNr")

        tab = h5.create_table(path, name, ProteinCacheInfo, createparents=True, obj=cached_cnts)
        tab.colinstances["EntryNr"].create_csindex()

        json_off = cache_h5.get_node("/family_json/offset").read()
        json_off.sort(order="Fam")
        tab = h5.root.RootHOG.MetaData.read()
        for fam, off, length in json_off:
            if tab[fam - 1]["FamNr"] != fam:
                raise DBConsistencyError("table not properly ordered")
            tab[fam - 1]["FamDataJsonOffset"] = off
            tab[fam - 1]["FamDataJsonLength"] = length
        h5.root.RootHOG.MetaData.modify_column(column=tab["FamDataJsonOffset"], colname="FamDataJsonOffset")
        h5.root.RootHOG.MetaData.modify_column(column=tab["FamDataJsonLength"], colname="FamDataJsonLength")

        json_in_buf = cache_h5.root.family_json.buffer
        json_in_buf._f_copy(h5.root.RootHOG, "JsonBuffer", expectedrows=len(json_in_buf))



from unittest import mock, SkipTest
import time
import os
from .test_db import TestWithDbInstance
from pyoma.browser.search import XRefSearch, search


class TestWithBigOmaDB(TestWithDbInstance):
    db_file_name = "All.Jul2023/data/OmaServer.h5"

    @classmethod
    def setUpClass(cls):
        with mock.patch.dict(os.environ, {"PYOMA_DB_PATH": "/Users/adriaal/Downloads"}):
            try:
                super().setUpClass()
            except IOError:
                raise SkipTest("Big database '{}' is not available".format(cls.db_file_name))


class XRefSearchTest(TestWithBigOmaDB):
    def test_existing_example(self):
        for query in ("TP53", "INS"):
            with self.subTest(query=query):
                s = XRefSearch(self.db, query, max_matches=80)
                hits = 0
                for p in s.search_entries():
                    for source, val in p.xref_data.items():
                        for el in val:
                            self.assertIn(query.lower(), el.lower())
                            hits += 1
                self.assertGreater(hits, 0)

    def test_combined_terms_failing(self):
        terms = [XRefSearch(self.db, term) for term in ("blue-light", "photoreceptor")]
        hits = 0
        for p in search(terms).entries:
            for source, val in p.xref_data.items():
                for el in val:
                    self.assertIn("blue-light", el.lower())
            hits += 1
        self.assertGreater(hits, 0)


if __name__ == "__main__":
    import unittest

    unittest.main()
